REM file: Whatis - DNDBBS Utility. PD v2.0a r3.0a. updated 02/20/2024

REM type: Recursive Descent Expression Parser.

REM compile vbdos:
REM  BC Whatis/ah/s/e/o/g3;
REM  BC Whatis1/ah/s/e/o/g3;
REM  BC Whatis2/ah/s/e/o/g3;
REM  LINK Whatis+Whatis1+Whatis2,,,vbdos/e;

' Bios Parameter Block Structure
TYPE BPBtype
 Level AS STRING * 2
 Serial AS STRING * 4
 Label AS STRING * 11
 System AS STRING * 8
END TYPE

' DOS Data Transfer Area
TYPE DTAtype
 Drive AS STRING * 1
 SearchTemplate AS STRING * 11
 SearchAttr AS STRING * 1
 EntryCount AS STRING * 2
 ClusterNumber AS STRING * 2
 Reserved AS STRING * 4
 FileAttr AS STRING * 1
 Filetime AS STRING * 2
 Filedate AS STRING * 2
 FileSize AS STRING * 4
 ASCIIZfilename AS STRING * 13
END TYPE

' interrupt structure
TYPE RegTypeX
 AX AS INTEGER
 BX AS INTEGER
 CX AS INTEGER
 DX AS INTEGER
 BP AS INTEGER
 SI AS INTEGER
 DI AS INTEGER
 Flags AS INTEGER
 DS AS INTEGER
 ES AS INTEGER
END TYPE

' share with whatis1
DECLARE SUB Vlabel (C$, C%)
DECLARE SUB Vserial (C$, C%)
DECLARE SUB Vtype (C$, C%)
DECLARE SUB InterruptX (N AS INTEGER, I AS RegTypeX, O AS RegTypeX)

COMMON SHARED Out2 AS STRING
COMMON SHARED Out3 AS STRING
COMMON SHARED Max.Row AS INTEGER
COMMON SHARED ControlBreak AS INTEGER
COMMON SHARED StdinEnabled AS INTEGER

' declare interrupt functions
DIM SHARED InregsX AS RegTypeX, OutregsX AS RegTypeX

' dta variables and structure declaration
DIM SHARED BASIC.DTA.SEG AS INTEGER
DIM SHARED BASIC.DTA.OFF AS INTEGER
DIM SHARED BPBfile AS BPBtype
DIM SHARED DTAfile AS DTAtype

' filename variables
DIM SHARED Drive.Search AS STRING * 1
DIM SHARED ASCIIZ AS STRING * 260
DIM SHARED Default.Net AS STRING * 16
DIM SHARED Current.Directory AS STRING
DIM SHARED LINUX AS INTEGER
DIM SHARED Pipe.Buffer AS STRING * 1

' function declarations
DECLARE FUNCTION ClearBuffer% ()
DECLARE FUNCTION KeyIS ()
DECLARE FUNCTION KeyboardChar$ ()
DECLARE FUNCTION KeyboardLine$ (V%)
DECLARE FUNCTION SearchLine$ (V%)
DECLARE FUNCTION STRIM$ (V$)
DECLARE FUNCTION XTRIM$ (V$)
DECLARE FUNCTION ParseSize$ (T#)
DECLARE FUNCTION PI# ()
DECLARE FUNCTION Timeit! ()

' subroutine declarations
DECLARE SUB Arith (Token.Parsed$, Temp#, Temp2#)
DECLARE SUB Assignment1 (S$, S%)
DECLARE SUB Assignment2 (S$, S1%, S2%)
DECLARE SUB ClearStatusLine ()
DECLARE SUB DisplayStatusLine (x%, Y%)
DECLARE SUB Enter.Equate ()
DECLARE SUB Equate (Temp#)
DECLARE SUB ExitProgram ()
DECLARE SUB Get.Machine.Name (x$)
DECLARE SUB Get.Token ()
DECLARE SUB Get.Token2 (Token.Exists%)
DECLARE SUB Get.Token3 ()
DECLARE SUB Get.Token4 ()
DECLARE SUB InstrSUB (Var%, Var1$, Var2$)
DECLARE SUB Out.File (V$, V%)
DECLARE SUB Parse.Alphabetic1 (Temp#)
DECLARE SUB Parse.Alphabetic2 (Temp#)
DECLARE SUB Parse.Alphabetic3 (Temp#)
DECLARE SUB ParseFN (Temp#)
DECLARE SUB ParseFZ (Temp#)
DECLARE SUB Parse.Numeric (Temp#)
DECLARE SUB Parse.Quoted ()
DECLARE SUB Parse1 (Temp#)
DECLARE SUB Parse2 (Temp#)
DECLARE SUB Parse3 (Temp#)
DECLARE SUB Parse4 (Temp#)
DECLARE SUB Parse5 (Temp#)
DECLARE SUB Parse6 (Temp#)
DECLARE SUB Parse7 (Temp#)
DECLARE SUB Parse7a (Temp#)
DECLARE SUB Parse8 (Temp#)
DECLARE SUB Parse9 (Temp#)
DECLARE SUB Read.Command (C$)
DECLARE SUB Read.Rows (Var%)
DECLARE SUB Remove.Spaces ()
DECLARE SUB StatusLine ()
DECLARE SUB FormatForm (T$, T#)
DECLARE SUB Swap.Data ()
DECLARE SUB Cryptograph (Var$)
DECLARE SUB Suffix (Var#, Var3$)
DECLARE SUB Save.Screen2 (Var%)
DECLARE SUB Get.Bits (Temp#)
DECLARE SUB Check.Drive (V%, T#)
DECLARE SUB Capitalize (Var3$)
DECLARE SUB CD.Directory ()
DECLARE SUB MD.Directory ()
DECLARE SUB RD.Directory ()
DECLARE SUB Name.Expression ()

DECLARE SUB Defseg.Function ()
DECLARE SUB Absolute.Function ()
DECLARE SUB Poke.Value ()
DECLARE SUB Output.Port ()
DECLARE SUB Wait.Port ()
DECLARE SUB Filedate (V$)
DECLARE SUB Filetime (V$)
DECLARE SUB GetFilesize (T#)
DECLARE SUB More.Prompt (I$, M$, O$)

DECLARE SUB Read.Stdin ()
DECLARE SUB Parse.Command.Line (V$)
DECLARE SUB Strip.Semicolon (V$)
DECLARE SUB List.Drives (VarQ%)
DECLARE SUB List.Files (VarQ%)
DECLARE SUB List.Dirs (VarQ%)
DECLARE SUB Playwav (Tmp1%, Tmp1$)

' set startup declarations
DEFINT A-Z
REM $DYNAMIC

' increase stack for recursion.
STACK 16384

' constant declarations
CONST FalseD = 0#
CONST TrueD = -1#
CONST False = 0
CONST True = -1
CONST Nul = ""
CONST Null$ = "<nul>"
CONST PrimeKey = 29
CONST MorePrompt = "More(y/n/c)?"
CONST MorePrompt2 = "More(y/n)?"
CONST MorePrompt3 = "-more(y/n/q/c/?)-"
CONST MorePrompt4 = "-more-"
CONST NodeX = ""

CONST Editor$ = "EDIT.COM"
CONST Interpreter$ = "QBASIC.EXE"

CONST Version$ = "v2.0a"
CONST Release$ = "r3.0a"

' maximum constant values (can be changed)
CONST Max.Arrays = 256
CONST Max.History = 10
CONST Max.Recurse = 16
CONST Max.FNs = 26
CONST Max.FN.Length = 1024

' define color values
CONST Black = 0
CONST Blue = 1
CONST Plain = 7
CONST Green = 10
CONST Cyan = 11
CONST Red = 12
CONST Magenta = 13
CONST Yellow = 14
CONST White = 15

' global variables
DIM SHARED Assign AS INTEGER
DIM SHARED DataError AS INTEGER
DIM SHARED Debug AS INTEGER
DIM SHARED Hidden AS INTEGER
DIM SHARED Last.Token AS INTEGER
DIM SHARED Out4 AS STRING
DIM SHARED Quote AS STRING * 1
DIM SHARED Recurse AS INTEGER
DIM SHARED Strng AS STRING
DIM SHARED ToggleForm AS INTEGER
DIM SHARED Timing AS INTEGER
DIM SHARED Token AS INTEGER
DIM SHARED Token.Index AS INTEGER
DIM SHARED Token.List AS STRING
DIM SHARED White.Space AS STRING

' savescreen arrays
DIM SHARED TempX AS INTEGER
DIM SHARED Temp.ArrayY(1) AS INTEGER
DIM SHARED Temp.ArrayZ(1) AS INTEGER

' data areas
DIM SHARED Arrays(0 TO 26, 1) AS DOUBLE ' variable arrays A() to Z()
DIM SHARED Definitions(1) AS STRING ' DEF FN functions
DIM SHARED Strngs(1) AS STRING ' string variables A$ to Z$
DIM SHARED Variables(1) AS DOUBLE ' variables A to Z

' multi-dimensional command history arrays
DIM SHARED History.Count2(1) AS INTEGER
DIM SHARED History2(1 TO 3, 1) AS STRING * 256
Dim Shared EndOfHistoryFlag(1) As Integer

' multi-dimensional command history variable
DIM SHARED Temp0 AS INTEGER

' program variables
REDIM Arrays(0 TO 26, 0 TO Max.Arrays) AS DOUBLE
REDIM Definitions(0 TO Max.FNs) AS STRING
REDIM Strngs(0 TO 26) AS STRING
REDIM Variables(0 TO 26) AS DOUBLE

' initialize multi-dimensional command history arrays
REDIM History2(1 TO 3, 1 TO Max.History) AS STRING * 256
REDIM History.Count2(1 TO 3) AS INTEGER
Redim EndOfHistoryFlag(1 To 3) As Integer

' reset multi-dimensional command history arrays
FOR VarQ1 = 1 TO 3
    EndOfHistoryFlag(VarQ1) = 0
    History.Count2(VarQ1) = 0
    FOR VarQ2 = 1 TO Max.History
	History2(VarQ1, VarQ2) = Nul
    NEXT
NEXT

' store basic dta.
InregsX.AX = &H2F00
CALL InterruptX(&H21, InregsX, OutregsX)
BASIC.DTA.SEG = OutregsX.ES
BASIC.DTA.OFF = OutregsX.BX

' store linux variable
IF ENVIRON$("LINUX") <> "" THEN
   LINUX = True
END IF

' store current directory
Current.Directory = CURDIR$
'IF RIGHT$(Current.Directory, 1) <> "\" THEN
'   Current.Directory = Current.Directory + "\"
'END IF

' declare global error routine
ON ERROR GOTO Error.Routine

' initialize some variables
Quote$ = CHR$(34)
Token.List = " -+*/\^()[]{}<>=|&!%~?:#@`;," + Quote$ ' cannot be changed
White.Space = CHR$(32) + CHR$(9) ' can be changed

' check command line exists
IF COMMAND$ = "" THEN GOTO StartLoop

' check override command line
IF ParseLine THEN
   Command.Line$ = COMMAND$
ELSE
   ' read lowercase command line from PSP
   CALL Read.Command(Command.Line$)
END IF

' check command line
Command.Line$ = LTRIM$(Command.Line$)
Command.Line$ = RTRIM$(Command.Line$)
IF LEFT$(Command.Line$, 1) = "@" THEN
   Filename$ = MID$(Command.Line$, 2)
   IF DIR$(Filename$) <> "" THEN
      x = FREEFILE
      OPEN Filename$ FOR INPUT AS #x
      Command.Line$ = Nul
      DO
	 IF EOF(x) THEN EXIT DO
	 LINE INPUT #x, Var$
	 Var$ = LTRIM$(RTRIM$(Var$))
	 IF LEN(Var$) THEN
	    Command.Line$ = Command.Line$ + ";" + Var$
	 END IF
      LOOP
   END IF
   CLOSE #x
   IF LEFT$(Command.Line$, 1) = ";" THEN
      Command.Line$ = MID$(Command.Line$, 2)
   END IF
END IF
IF LEN(Command.Line$) THEN
    V$ = Command.Line$
    CALL Parse.Command.Line(V$)
    ' restore current drive/directory
    CHDRIVE Current.Directory
    CHDIR Current.Directory
Error.Resume1:
 COLOR Plain, Black
 END
END IF

' reset error flow
StartLoop:
DataError = True

' run program from stdin
CALL Read.Stdin

' display statusline at max.row
CALL Read.Rows(Max.Row)
CALL Save.Screen2(0)
CALL StatusLine
GOSUB Display.Header
Error.Resume2:

' keyboard input loop
DO
    COLOR Yellow, Black
    LOCATE , 1, 1
    PRINT ">";
    LOCATE , , 1
    ' get input line
    Temp0 = 3
    Out2 = KeyboardLine$(-1)
    PRINT
StartInput:
    ' parse input line
    SELECT CASE UCASE$(Out2)
    CASE "?"
       GOSUB Display.Header
    CASE "TIMING"
       Timing = NOT Timing
       IF Timing THEN
	  PRINT "Timing on."
       ELSE
	  PRINT "Timing off."
       END IF
    CASE "TOGGLE"
       ToggleForm = NOT ToggleForm
       IF ToggleForm THEN
	  PRINT "Toggled on."
       ELSE
	  PRINT "Toggled off."
       END IF
    CASE "QUIT"
       EXIT DO
    CASE "ASCII"
       CALL Out.File("ASCII.TXT", 0)
    CASE "HELP"
       CALL Out.File("WHATIS.DOC", -1)
    CASE "WHATIS"
       CALL Out.File("WHATIS.TXT", -1)
    CASE "CLEAR"
       FOR V = 1 TO Max.History
	  History2(Temp0, V) = Nul
       NEXT
       COLOR White, Black
       PRINT "Command history cleared."
    CASE "LIST"
       COLOR White, Black
       PRINT "Command history:"
       FOR V = 1 TO Max.History
	  V$ = RTRIM$(History2(Temp0, V))
	  PRINT V; " ";
	  IF LEN(V$) THEN
	     FOR VarX = 1 TO LEN(V$)
		PRINT MID$(V$, VarX, 1);
	     NEXT
	     PRINT
	  ELSE
	     PRINT Null$
	  END IF
       NEXT
       COLOR Yellow, Black
       CALL More.Prompt(MorePrompt4, "yn" + CHR$(13), Output.Char$)
    ' these functions are unstable.
    CASE "DRIVES"
       CALL List.Drives(0)
    CASE "DIRS"
       CALL List.Dirs(-1)
    CASE "FILES"
       CALL List.Files(-1)
    CASE "SEARCH"
       Out2 = SearchLine$(-1)
       PRINT
       GOTO StartInput
    CASE ELSE
       ' calculate line
       CALL Remove.Spaces
       Recurse = 0
       Hidden = False
       StartTime# = Timeit!
       CALL Enter.Equate
       IF Assign THEN
	  IF Hidden = 0 THEN
	     COLOR White, Black
	     PRINT "Value assigned."
	  END IF
       END IF
       IF Timing THEN
	  RunTime# = Timeit! - StartTime#
	  IF RunTime# < 0# THEN RunTime# = RunTime# + 86400#
	  PRINT "Time elapsed:"; RunTime#
       END IF
       Var1 = CSRLIN
       Var2 = POS(0)
       CALL DisplayStatusLine(Var1, Var2)
    END SELECT
LOOP

' end program
COLOR White, Black
PRINT "Press <enter> to return to DOS:";
DO
    Input.Char$ = KeyboardChar$
    IF Input.Char$ = CHR$(13) THEN
	EXIT DO
    END IF
LOOP
PRINT

' clear statusline and end program
CALL ExitProgram
END

Display.Header:
' display header
COLOR White, Black
PRINT "Whatis " + Version$ + " " + Release$ + ": Recursive Descent Expression Parser;"
PRINT "  Functions:"
PRINT "    'quit' return to system"
PRINT "    'help' display whatis.doc"
PRINT "    'list' to list history"
PRINT "    'clear' to clear history"
PRINT "    'ascii' display ascii.txt"
PRINT "    'search' search history"
PRINT "    'timing' toggle timer"
PRINT "    'toggle' switch to format mode"
PRINT "    'whatis' display whatis.txt"
RETURN

StatusLineTrap:
 Var1 = CSRLIN
 Var2 = POS(0)
 CALL DisplayStatusLine(Var1, Var2)
 RETURN

' main error routine
Error.Routine:
IF DataError = False THEN
    PRINT
END IF
COLOR Red, Black
Error.Mode = ERR
SELECT CASE Error.Mode
    ' standard errors.
    CASE 7
	PRINT "Out of memory."
    CASE 9
	PRINT "Subscript out of range."
    CASE 14
	PRINT "Out of string space."
    CASE 53
	PRINT "File not found."
    CASE 57
	PRINT "Media error."
    CASE 61
	PRINT "Disk full."
    CASE 64
        PRINT "Bad file name."
    CASE 68
	PRINT "Device unavailable."
    CASE 70
	PRINT "Permission denied."
    CASE 71
	PRINT "Disk not ready."
    CASE 75, 76
	PRINT "Path not found."
	' extended errors.
    CASE 5
	PRINT "Error #001: Illegal function call."
    CASE 6
	PRINT "Error #002: Overflow."
    CASE 9
	PRINT "Error #003: Subscript out of range."
    CASE 11
	PRINT "Error #004: Division by zero."
    CASE 92
	PRINT "Error #005: Whatis: "; Strng
    CASE 130
	PRINT "Error #006: Bad numeric specifier."
    CASE 140
	PRINT "Error #007: Bad exponent specifier."
    CASE 145
	PRINT "Error #008: Bad DEFFN declaration."
    CASE 150
	PRINT "Error #009: Bad swap statement."
    CASE ELSE
	'PRINT "Critical error:" + STR$(Error.Mode) + " in Whatis."
	PRINT "Error #000: Error"; Error.Mode
END SELECT
COLOR White, Black
LOCATE , , 1
Error.Exit1:
IF DataError = False THEN
    RESUME Error.Resume1
ELSE
    RESUME Error.Resume2
END IF
END

REM end of main program

SUB Absolute.Function
 CALL Get.Token3
 CALL Get.Token
 CALL Parse1(Temp#)
 CALL Absolute(CINT(Temp#))
 CALL Get.Token4
 Last.Token = True
END SUB

' routine accepts an operation and performs on two values, string or numeric
SUB Arith (Token.Parsed$, Temp#, Temp2#)
IF Last.Token THEN
    SELECT CASE Token.Parsed$
	CASE "-"
	    Temp# = Temp# - Temp2#
	CASE "+"
	    Temp# = Temp# + Temp2#
	CASE "/"
	    Temp# = Temp# / Temp2#
	CASE "\"
	    Temp# = Temp# \ Temp2#
	CASE "*"
	    Temp# = Temp# * Temp2#
	CASE "^"
	    Temp# = Temp# ^ Temp2#
	CASE "||"
	    Temp# = Temp# * 10#
	CASE "##"
	    Temp# = Temp# \ 10#
	CASE "^^"
	    Temp# = Temp# ^ 10#
	CASE "^=", "^+="
	    Temp# = Temp# ^ Temp2#
	CASE "^-="
	    Temp# = Temp# ^ (-Temp2#)
	CASE "<"
	    Temp# = (Temp# < Temp2#)
	CASE ">"
	    Temp# = (Temp# > Temp2#)
	CASE "<<"
	    Temp# = Temp# * 2#
	CASE ">>"
	    Temp# = Temp# \ 2#
	CASE "++"
	    Temp# = Temp# + 1#
	CASE "--"
	    Temp# = Temp# - 1#
	CASE "**"
	    Temp# = Temp# ^ 2#
	CASE "//"
	    Temp# = SQR(Temp#)
	CASE "=="
	    Temp# = Temp# + 0#
	CASE "="
	    Temp# = (Temp# = Temp2#)
	CASE "<="
	    Temp# = (Temp# <= Temp2#)
	CASE ">="
	    Temp# = (Temp# >= Temp2#)
	CASE "<>"
	    Temp# = (Temp# <> Temp2#)
	CASE "|"
	    Temp# = Temp# OR Temp2#
	CASE "&"
	    Temp# = Temp# AND Temp2#
	CASE "%"
	    Temp# = Temp# MOD Temp2#
	CASE "~"
	    Temp# = Temp# XOR Temp2#
	CASE "?"
	    Temp# = Temp# IMP Temp2#
	CASE ":"
	    Temp# = Temp# EQV Temp2#
	CASE "#"
	    Temp# = NOT (Temp# OR Temp2#)
	CASE "@"
	    Temp# = NOT (Temp# IMP Temp2#)
	CASE "`"
	    Temp# = NOT (Temp# AND Temp2#)
	CASE ELSE
	   Strng = "<unknown token>"
	   ERROR 92
    END SELECT
    EXIT SUB
END IF
SELECT CASE Token.Parsed$
    CASE "-"
	IF RIGHT$(Out3, LEN(Out4)) = Out4 THEN
	    Out3 = LEFT$(Out3, LEN(Out3) - LEN(Out4))
	END IF
    CASE "+", "*"
	Out3 = Out3 + Out4
    CASE "/" ' remove one instance
	IF LEN(Out3) > False AND LEN(Out4) > False THEN
	    Imbedded = INSTR(Out3, Out4)
	    IF Imbedded THEN
		Out3 = LEFT$(Out3, Imbedded - 1) + MID$(Out3, Imbedded + LEN(Out4))
	    END IF
	END IF
    CASE "\" ' remove all instances
	IF LEN(Out3) > False AND LEN(Out4) > False THEN
	    Imbedded = INSTR(Out3, Out4)
	    WHILE Imbedded
		Out3 = LEFT$(Out3, Imbedded - 1) + MID$(Out3, Imbedded + LEN(Out4))
		Imbedded = INSTR(Out3, Out4)
	    WEND
	END IF
    CASE "<"
	Last.Token = True
	Temp# = Out3 < Out4
    CASE ">"
	Last.Token = True
	Temp# = Out3 > Out4
    CASE "="
	Last.Token = True
	Temp# = Out3 = Out4
    CASE "<="
	Last.Token = True
	Temp# = Out3 <= Out4
    CASE ">="
	Last.Token = True
	Temp# = Out3 >= Out4
    CASE "<>", "#"
	Last.Token = True
	Temp# = Out3 <> Out4
    CASE "=="
	Last.Token = True
	Temp# = Temp#
    CASE ELSE
	Strng = "<unknown token>"
	ERROR 92
END SELECT
END SUB

SUB Assignment1 (S$, S%)
IF LEFT$(S$, 2) = "<<" THEN
    Assign = True
    Strng = MID$(S$, 3)
    IF Strng = Nul THEN
	Variables(S%) = Variables(S%) * 2#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = ">>" THEN
    Assign = True
    Strng = MID$(S$, 3)
    IF Strng = Nul THEN
	Variables(S%) = Variables(S%) \ 2#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "||" THEN
    Assign = True
    Strng = MID$(S$, 3)
    IF Strng = Nul THEN
	Variables(S%) = Variables(S%) * 10#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "##" THEN
    Assign = True
    Strng = MID$(S$, 3)
    IF Strng = Nul THEN
	Variables(S%) = Variables(S%) \ 10#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "^^" THEN
    Assign = True
    Strng = MID$(S$, 3)
    IF Strng = Nul THEN
	Variables(S%) = Variables(S%) ^ 10#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "--" THEN
    Assign = True
    Strng = MID$(S$, 3)
    IF Strng = Nul THEN
	Variables(S%) = Variables(S%) - 1#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "++" THEN
    Assign = True
    Strng = MID$(S$, 3)
    IF Strng = Nul THEN
	Variables(S%) = Variables(S%) + 1#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "==" THEN
    Assign = True
    Strng = MID$(S$, 3)
    IF Strng = Nul THEN
	Variables(S%) = Variables(S%) + 0#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "**" THEN
    Assign = True
    Strng = MID$(S$, 3)
    IF Strng = Nul THEN
	Variables(S%) = Variables(S%) ^ 2#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "//" THEN
    Assign = True
    Strng = MID$(S$, 3)
    IF Strng = Nul THEN
	Variables(S%) = SQR(Variables(S%))
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "-=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    CALL Equate(T#)
    IF Last.Token = True THEN
	Assign = True
	Variables(S%) = Variables(S%) - T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "+=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    CALL Equate(T#)
    IF Last.Token = True THEN
	Assign = True
	Variables(S%) = Variables(S%) + T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "/=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    CALL Equate(T#)
    IF Last.Token = True THEN
	Assign = True
	Variables(S%) = Variables(S%) / T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "\=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    CALL Equate(T#)
    IF Last.Token = True THEN
	Assign = True
	Variables(S%) = Variables(S%) \ T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "*=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    CALL Equate(T#)
    IF Last.Token = True THEN
	Assign = True
	Variables(S%) = Variables(S%) * T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "%=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    CALL Equate(T#)
    IF Last.Token = True THEN
	Assign = True
	Variables(S%) = Variables(S%) MOD T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "|=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    CALL Equate(T#)
    IF Last.Token = True THEN
	Assign = True
	Variables(S%) = Variables(S%) OR T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "&=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    CALL Equate(T#)
    IF Last.Token = True THEN
	Assign = True
	Variables(S%) = Variables(S%) AND T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "~=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    CALL Equate(T#)
    IF Last.Token = True THEN
	Assign = True
	Variables(S%) = Variables(S%) XOR T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "?=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    CALL Equate(T#)
    IF Last.Token = True THEN
	Assign = True
	Variables(S%) = Variables(S%) IMP T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = ":=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    CALL Equate(T#)
    IF Last.Token = True THEN
	Assign = True
	Variables(S%) = Variables(S%) EQV T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "#=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    CALL Equate(T#)
    IF Last.Token = True THEN
	Assign = True
	Variables(S%) = NOT (Variables(S%) OR T#)
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "@=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    CALL Equate(T#)
    IF Last.Token = True THEN
	Assign = True
	Variables(S%) = NOT (Variables(S%) IMP T#)
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "`=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    CALL Equate(T#)
    IF Last.Token = True THEN
	Assign = True
	Variables(S%) = NOT (Variables(S%) AND T#)
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 1) = "=" THEN
    Out2 = MID$(S$, 2)
    Out2 = LTRIM$(Out2)
    CALL Equate(T#)
    IF Last.Token = True THEN
	Assign = True
	Variables(S%) = T#
    END IF
    EXIT SUB
END IF
END SUB

SUB Assignment2 (S$, S1%, S2%)
IF LEFT$(S$, 2) = "<<" THEN
    Assign = True
    Strng = MID$(S$, 3)
    IF Strng = Nul THEN
	Arrays(S1%, S2%) = Arrays(S1%, S2%) * 2#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = ">>" THEN
    Assign = True
    Strng = MID$(S$, 3)
    IF Strng = Nul THEN
	Arrays(S1%, S2%) = Arrays(S1%, S2%) \ 2#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "||" THEN
    Assign = True
    Strng = MID$(S$, 3)
    IF Strng = Nul THEN
	Arrays(S1%, S2%) = Arrays(S1%, S2%) * 10#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "##" THEN
    Assign = True
    Strng = MID$(S$, 3)
    IF Strng = Nul THEN
	Arrays(S1%, S2%) = Arrays(S1%, S2%) \ 10#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "^^" THEN
    Assign = True
    Strng = MID$(S$, 3)
    IF Strng = Nul THEN
	Arrays(S1%, S2%) = Arrays(S1%, S2%) ^ 10#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "--" THEN
    Assign = True
    Strng = MID$(S$, 3)
    IF Strng = Nul THEN
	Arrays(S1%, S2%) = Arrays(S1%, S2%) - 1#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "++" THEN
    Assign = True
    Strng = MID$(S$, 3)
    IF Strng = Nul THEN
	Arrays(S1%, S2%) = Arrays(S1%, S2%) + 1#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "==" THEN
    Assign = True
    Strng = MID$(S$, 3)
    IF Strng = Nul THEN
	Arrays(S1%, S2%) = Arrays(S1%, S2%) + 0#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "**" THEN
    Assign = True
    Strng = MID$(S$, 3)
    IF Strng = Nul THEN
	Arrays(S1%, S2%) = Arrays(S1%, S2%) ^ 2#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "//" THEN
    Assign = True
    Strng = MID$(S$, 3)
    IF Strng = Nul THEN
	Arrays(S1%, S2%) = SQR(Arrays(S1%, S2%))
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "-=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    Token.Index = 1
    CALL Get.Token
    CALL Parse1(T#)
    IF Last.Token = True THEN
	Assign = True
	Arrays(S1%, S2%) = Arrays(S1%, S2%) - T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "+=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    Token.Index = 1
    CALL Get.Token
    CALL Parse1(T#)
    IF Last.Token = True THEN
	Assign = True
	Arrays(S1%, S2%) = Arrays(S1%, S2%) + T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "/=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    Token.Index = 1
    CALL Get.Token
    CALL Parse1(T#)
    IF Last.Token = True THEN
	Assign = True
	Arrays(S1%, S2%) = Arrays(S1%, S2%) / T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "\=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    Token.Index = 1
    CALL Get.Token
    CALL Parse1(T#)
    IF Last.Token = True THEN
	Assign = True
	Arrays(S1%, S2%) = Arrays(S1%, S2%) \ T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "*=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    Token.Index = 1
    CALL Get.Token
    CALL Parse1(T#)
    IF Last.Token = True THEN
	Assign = True
	Arrays(S1%, S2%) = Arrays(S1%, S2%) * T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "%=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    Token.Index = 1
    CALL Get.Token
    CALL Parse1(T#)
    IF Last.Token = True THEN
	Assign = True
	Arrays(S1%, S2%) = Arrays(S1%, S2%) MOD T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "|=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    Token.Index = 1
    CALL Get.Token
    CALL Parse1(T#)
    IF Last.Token = True THEN
	Assign = True
	Arrays(S1%, S2%) = Arrays(S1%, S2%) OR T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "&=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    Token.Index = 1
    CALL Get.Token
    CALL Parse1(T#)
    IF Last.Token = True THEN
	Assign = True
	Arrays(S1%, S2%) = Arrays(S1%, S2%) AND T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "~=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    Token.Index = 1
    CALL Get.Token
    CALL Parse1(T#)
    IF Last.Token = True THEN
	Assign = True
	Arrays(S1%, S2%) = Arrays(S1%, S2%) XOR T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "?=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    Token.Index = 1
    CALL Get.Token
    CALL Parse1(T#)
    IF Last.Token = True THEN
	Assign = True
	Arrays(S1%, S2%) = Arrays(S1%, S2%) IMP T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = ":=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    Token.Index = 1
    CALL Get.Token
    CALL Parse1(T#)
    IF Last.Token = True THEN
	Assign = True
	Arrays(S1%, S2%) = Arrays(S1%, S2%) EQV T#
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "#=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    Token.Index = 1
    CALL Get.Token
    CALL Parse1(T#)
    IF Last.Token = True THEN
	Assign = True
	Arrays(S1%, S2%) = NOT (Arrays(S1%, S2%) OR T#)
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "@=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    Token.Index = 1
    CALL Get.Token
    CALL Parse1(T#)
    IF Last.Token = True THEN
	Assign = True
	Arrays(S1%, S2%) = NOT (Arrays(S1%, S2%) IMP T#)
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 2) = "`=" THEN
    Out2 = MID$(S$, 3)
    Out2 = LTRIM$(Out2)
    Token.Index = 1
    CALL Get.Token
    CALL Parse1(T#)
    IF Last.Token = True THEN
	Assign = True
	Arrays(S1%, S2%) = NOT (Arrays(S1%, S2%) AND T#)
    END IF
    EXIT SUB
END IF
IF LEFT$(S$, 1) = "=" THEN
    Out2 = MID$(S$, 2)
    Out2 = LTRIM$(Out2)
    Token.Index = 1
    CALL Get.Token
    CALL Parse1(T#)
    IF Last.Token = True THEN
	Assign = True
	Arrays(S1%, S2%) = T#
    END IF
END IF
END SUB

SUB Capitalize (Var3$)
Var3$ = RTRIM$(LTRIM$(Var3$))
IF Var3$ = Nul THEN
   EXIT SUB
END IF
MID$(Var3$, 1, 1) = UCASE$(MID$(Var3$, 1, 1))
Var = INSTR(Var3$, " ")
IF Var < LEN(Var3$) THEN
   DO WHILE Var
      MID$(Var3$, Var + 1, 1) = UCASE$(MID$(Var3$, Var + 1, 1))
      Var = INSTR(Var + 1, Var3$, " ")
   LOOP
END IF
END SUB

SUB CD.Directory
 CALL Get.Token3
 CALL Get.Token
 CALL Parse1(Temp#)
 CALL Get.Token4
 Hidden = True
 CHDIR Out3
 Last.Token = False
END SUB

' check drive exists
SUB Check.Drive (V, T#)
 ON LOCAL ERROR GOTO Disk.Error
 ' check disk space.
 InregsX.AX = &H3600
 InregsX.DX = V
 CALL InterruptX(&H21, InregsX, OutregsX)
 IF OutregsX.AX = &HFFFF THEN
    T# = 0#
 ELSE
    T# = -1#
 END IF
 EXIT SUB
Disk.Error:
 T# = 0#
 EXIT SUB
END SUB

' tests drive
SUB Check.Drive2 (V, T#)
 ON LOCAL ERROR GOTO Disk.Error2
 ' get volume info
 InregsX.AX = &H6900
 InregsX.BX = V
 InregsX.DS = VARSEG(BPBfile)
 InregsX.DX = VARPTR(BPBfile)
 CALL InterruptX(&H21, InregsX, OutregsX)
 ' check flag error
 IF (OutregsX.Flags AND &H1) = &H0 THEN
    T# = -1#
 ELSE
    T# = 0#
 END IF
 EXIT SUB
Disk.Error2:
 T# = 0#
 EXIT SUB
END SUB

' clears keyboard buffer
FUNCTION ClearBuffer
 DO
    IF KeyIS THEN
       Char$ = KeyboardChar$
    ELSE
       EXIT DO
    END IF
 LOOP
 ClearBuffer = True
END FUNCTION

SUB ClearStatusLine
   LOCATE Max.Row, 1, 1
   COLOR Plain, Black
   PRINT SPACE$(80);
   LOCATE Max.Row - 1, 1, 1
END SUB

SUB Cryptograph (Var$)
 ON LOCAL ERROR GOTO Error.Trap54
 Var1 = LEN(Var$)
 IF Var1 = False THEN
    EXIT SUB
 END IF
 FOR Var2 = False TO Var1 - 1
    VarZ = ASC(MID$(Var$, Var2 + 1, 1))
    VarZ = VarZ XOR Var1
    IF Var2 = False THEN
       VarZ = VarZ XOR PrimeKey
    ELSE
       VarZ = VarZ XOR Var2
       VarZ = VarZ XOR ASC(MID$(Var$, Var2, 1))
    END IF
    MID$(Var$, Var2 + 1, 1) = CHR$(VarZ)
 NEXT
 IF Var1 = 1 THEN
    EXIT SUB
 END IF
 FOR Var2 = Var1 - 1 TO 0 STEP -1
    VarZ = ASC(MID$(Var$, Var2 + 1, 1))
    VarZ = VarZ XOR Var1
    IF Var2 = Var1 - 1 THEN
       VarZ = VarZ XOR PrimeKey
    ELSE
       VarZ = VarZ XOR Var2
       VarZ = VarZ XOR ASC(MID$(Var$, Var2 + 2, 1))
    END IF
    MID$(Var$, Var2 + 1, 1) = CHR$(VarZ)
 NEXT
Error.Resume54:
 EXIT SUB
Error.Trap54:
 RESUME Error.Resume54
END SUB

SUB Defseg.Function
 Var$ = MID$(Out2, 7)
 Var$ = LTRIM$(Var$)
 IF Var$ = Nul THEN
    DEF SEG ' restore segment
    Last.Token = True
    EXIT SUB
 END IF
 CALL Get.Token3
 CALL Get.Token
 CALL Parse1(Temp#)
 DEF SEG = Temp# ' set segment
 CALL Get.Token4
 Last.Token = True
END SUB

SUB DisplayStatusLine (Var1, Var2)
REM Var1, Var2 = Cursor position X, Y
IF StdinEnabled THEN EXIT SUB
S$ = Version$ + " " + Release$ + " " + FORMAT$(Now, "mm-dd-yyyy") + " " + FORMAT$(Now, "hh:mm:ss")
IF Debug THEN
   Var2# = FRE(-1) + FRE("A") ' far+near segments
   Var2$ = FORMAT$(Var2#, "#,##0.00;;")
   S$ = "- Free RAM: " + Var2$ + " bytes."
END IF
LOCATE Max.Row, 1, 1
COLOR White, Blue
S$ = "Whatis utility " + S$
S$ = LEFT$(S$, 80)
S$ = S$ + SPACE$(80 - LEN(S$))
PRINT S$;
COLOR Yellow, Black
LOCATE Var1, Var2, 1
END SUB

' beginning of the equation parser
SUB Enter.Equate
REM entry to expression parser, each following parser is higher precedence,
REM this routine called by higher precedence routines recursively.

REM precedence of operators:
REM   !, +, -  Not, Unary plus, Unary minus
REM   <<, >>, --, ++, **, //, ||, ##, ^^  Dual unary
REM   ^, ^=, ^-=, ^+=   Power/Negative power
REM   *, /     Multiplication, Division
REM   \        Integer Division
REM   %        Modulo
REM   +, -     Addition, Subtraction
REM   =, <, >, <>, >=, <=   Relational
REM   |, &, ~, ?, :, #, @, `  Logical (or, and, xor, imp, eqv, nor, non, xan)

Strng = Nul
Out2 = LTRIM$(Out2)
IF Out2 = Nul THEN
    EXIT SUB
END IF
Store.Input$ = Out2
Assign = False
GOSUB Assignment
IF Assign THEN
    EXIT SUB
END IF
Out2 = Store.Input$
D$ = UCASE$(Out2)
IF LEFT$(D$, 4) = "SWAP" THEN
    CALL Swap.Data
    EXIT SUB
END IF
IF LEFT$(D$, 3) = "MID" THEN
    GOSUB Assign.Mid.String
    IF Assign THEN
	EXIT SUB
    END IF
END IF
IF LEFT$(D$, 4) = "LEFT" THEN
    GOSUB Assign.Left.String
    IF Assign THEN
	EXIT SUB
    END IF
END IF
IF LEFT$(D$, 5) = "RIGHT" THEN
    GOSUB Assign.Right.String
    IF Assign THEN
	EXIT SUB
    END IF
END IF
IF LEFT$(D$, 5) = "DEFFN" THEN
    GOSUB Define.Function
    IF Assign THEN
	EXIT SUB
    END IF
END IF
Out2 = Store.Input$
Last.Token = False
Out3 = Nul
Out4 = Nul
Temp# = False
Token.Index = 1
CALL Get.Token
CALL Parse1(T#)

' parse leftover tokens
WHILE LEN(Strng)
    CALL Parse1(x#)
    ' toss away any unwanted tokens
    IF Last.Token = 3 THEN
	Strng = "<extra token>"
	ERROR 92
    END IF
WEND

' display output result
IF Hidden THEN
   EXIT SUB
END IF
COLOR White, Black
PRINT "Equals: ";
IF Last.Token = False THEN
    IF Out3 = Nul THEN
       COLOR Red, Black
       PRINT Null$
       COLOR White, Black
    ELSE
       COLOR Green, Black
       PRINT Out3
       COLOR White, Black
    END IF
ELSE
    COLOR Magenta, Black
    IF ToggleForm THEN
       CALL FormatForm(T$, T#)
       PRINT T$
    ELSE
       PRINT LTRIM$(STR$(T#))
    END IF
    COLOR White, Black
END IF
EXIT SUB

Assignment:
Assign = False
Temp1$ = UCASE$(LEFT$(Out2, 1))
IF Temp1$ = "?" THEN
    Temp1$ = MID$(Out2, 2)
    Temp2$ = LEFT$(Temp1$, 1)
    IF Temp2$ = "=" THEN
	Temp1$ = MID$(Temp1$, 2)
	Out2 = Temp1$
	' calculates expression w/o output
	Store.Input$ = Out2
	CALL Remove.Spaces
	Recurse = 0
	Hidden = True
	CALL Enter.Equate
	'Assign = True
	EXIT SUB
    END IF
END IF
IF Temp1$ >= "A" AND Temp1$ <= "Z" THEN
    Variable = ASC(Temp1$) - 64
    Temp1$ = MID$(Out2, 2)
    Temp1$ = LTRIM$(Temp1$)
    CALL Assignment1(Temp1$, Variable)
    IF Assign THEN
	GOSUB Assign.End
	RETURN
    END IF
    IF LEFT$(Temp1$, 1) = "$" THEN
	Temp1$ = MID$(Temp1$, 2)
	Temp1$ = LTRIM$(Temp1$)
	IF LEFT$(Temp1$, 1) = "(" THEN
	    Strng = "<bad string token>"
	    ERROR 92
	END IF
	IF LEFT$(Temp1$, 1) = "[" THEN
	    Strng = "<bad string token>"
	    ERROR 92
	END IF
	IF LEFT$(Temp1$, 1) = "{" THEN
	    Strng = "<bad string token>"
	    ERROR 92
	END IF
	IF LEFT$(Temp1$, 1) = "=" THEN
	    Out2 = MID$(Temp1$, 2)
	    Out2 = LTRIM$(Out2)
	    Token.Index = 1
	    CALL Get.Token
	    CALL Parse1(Temp3#)
	    IF Last.Token = False THEN
		Strngs(Variable) = Out3
		Assign = True
		GOSUB Assign.End
		RETURN
	    END IF
	END IF
    END IF
    IF LEFT$(Temp1$, 1) = "(" THEN
	Out2 = MID$(Temp1$, 2)
	Out2 = LTRIM$(Out2)
	Last.Token = False
	Token.Index = 1
	CALL Get.Token
	CALL Parse1(Temp3#)
	IF Strng <> ")" THEN
	   Strng = "<bad closing token>"
	   ERROR 92
	END IF
	IF Last.Token = True THEN
	    Element = CINT(Temp3#)
	    Temp1$ = MID$(Out2, Token.Index)
	    Temp1$ = LTRIM$(Temp1$)
	    CALL Assignment2(Temp1$, Variable, Element)
	    IF Assign THEN
		GOSUB Assign.End
		RETURN
	    END IF
	END IF
    END IF
    IF LEFT$(Temp1$, 1) = "[" THEN
	Out2 = MID$(Temp1$, 2)
	Out2 = LTRIM$(Out2)
	Last.Token = False
	Token.Index = 1
	CALL Get.Token
	CALL Parse1(Temp3#)
	IF Strng <> "]" THEN
	   Strng = "<bad closing token>"
	   ERROR 92
	END IF
	IF Last.Token = True THEN
	    Element = CINT(Temp3#)
	    Temp1$ = MID$(Out2, Token.Index)
	    Temp1$ = LTRIM$(Temp1$)
	    CALL Assignment2(Temp1$, Variable, Element)
	    IF Assign THEN
		GOSUB Assign.End
		RETURN
	    END IF
	END IF
    END IF
    IF LEFT$(Temp1$, 1) = "{" THEN
	Out2 = MID$(Temp1$, 2)
	Out2 = LTRIM$(Out2)
	Last.Token = False
	Token.Index = 1
	CALL Get.Token
	CALL Parse1(Temp3#)
	IF Strng <> "}" THEN
	   Strng = "<bad closing token>"
	   ERROR 92
	END IF
	IF Last.Token = True THEN
	    Element = CINT(Temp3#)
	    Temp1$ = MID$(Out2, Token.Index)
	    Temp1$ = LTRIM$(Temp1$)
	    CALL Assignment2(Temp1$, Variable, Element)
	    IF Assign THEN
		GOSUB Assign.End
		RETURN
	    END IF
	END IF
    END IF
END IF
RETURN

Assign.Mid.String:
Out2 = MID$(Out2, 4)
IF LEN(Out2) THEN
    Out2 = LTRIM$(Out2)
    Out2 = MID$(Out2, 3)
    V$ = UCASE$(LEFT$(Out2, 1))
    IF V$ >= "A" AND V$ <= "Z" THEN
	Variable = ASC(V$) - 64
	IF MID$(Out2, 2, 1) = "$" THEN
	    Out2 = MID$(Out2, 4)
	    Last.Token = False
	    Token.Index = 1
	    CALL Get.Token
	    CALL Parse1(T#)
	    Start = CINT(T#)
	    IF Strng = "," THEN
		CALL Get.Token
		CALL Parse1(T#)
		Length = CINT(T#)
		Out2 = MID$(Out2, Token.Index)
		Out2 = LTRIM$(Out2)
		IF LEFT$(Out2, 1) = "=" THEN
		    Out2 = MID$(Out2, 2)
		    Out2 = LTRIM$(Out2)
		    Token.Index = 1
		    CALL Get.Token
		    CALL Parse1(T#)
		    IF Last.Token = False THEN
			Assign = True
			IF Strngs(Variable) = Nul THEN
			    IF Start = 1 THEN
				IF LEN(Out3) = Length THEN
				    Strngs(Variable) = Out3
				    GOSUB Assign.End
				    RETURN
				END IF
			    END IF
			    Strng = "<bad string>"
			    ERROR 92
			    RETURN
			END IF
			MID$(Strngs(Variable), Start, Length) = Out3
			GOSUB Assign.End
			RETURN
		    END IF
		END IF
	    END IF
	END IF
    END IF
END IF
RETURN

Assign.Left.String:
Out2 = MID$(Out2, 5)
IF LEN(Out2) THEN
    Out2 = LTRIM$(Out2)
    Out2 = MID$(Out2, 3)
    V$ = UCASE$(LEFT$(Out2, 1))
    IF V$ >= "A" AND V$ <= "Z" THEN
	Variable = ASC(V$) - 64
	IF MID$(Out2, 2, 1) = "$" THEN
	    Out2 = MID$(Out2, 4)
	    Last.Token = False
	    Token.Index = 1
	    CALL Get.Token
	    CALL Parse1(T#)
	    Start = CINT(T#)
	    Out2 = MID$(Out2, Token.Index)
	    Out2 = LTRIM$(Out2)
	    IF LEFT$(Out2, 1) = "=" THEN
		Assign = True
		Out2 = MID$(Out2, 2)
		Out2 = LTRIM$(Out2)
		Token.Index = 1
		CALL Get.Token
		CALL Parse1(T#)
		IF Last.Token = False THEN
		    Assign = True
		    IF Strngs(Variable) = Nul THEN
			IF LEN(Out3) = Start THEN
			    Strngs(Variable) = Out3
			    GOSUB Assign.End
			    RETURN
			END IF
			Strng = "<bad string>"
			ERROR 92
			RETURN
		    END IF
		    MID$(Strngs(Variable), 1, Start) = Out3
		    GOSUB Assign.End
		    RETURN
		END IF
	    END IF
	END IF
    END IF
END IF
RETURN

Assign.Right.String:
Out2 = MID$(Out2, 6)
IF LEN(Out2) THEN
    Out2 = LTRIM$(Out2)
    Out2 = MID$(Out2, 3)
    V$ = UCASE$(LEFT$(Out2, 1))
    IF V$ >= "A" AND V$ <= "Z" THEN
	Variable = ASC(V$) - 64
	IF MID$(Out2, 2, 1) = "$" THEN
	    Out2 = MID$(Out2, 4)
	    Last.Token = False
	    Token.Index = 1
	    CALL Get.Token
	    CALL Parse1(T#)
	    Start = CINT(T#)
	    Out2 = MID$(Out2, Token.Index)
	    Out2 = LTRIM$(Out2)
	    IF LEFT$(Out2, 1) = "=" THEN
		Assign = True
		Out2 = MID$(Out2, 2)
		Out2 = LTRIM$(Out2)
		Token.Index = 1
		CALL Get.Token
		CALL Parse1(T#)
		Length = LEN(Strngs(Variable)) - Start + 1
		IF Last.Token = False THEN
		    Assign = True
		    IF Length > False THEN
			MID$(Strngs(Variable), Length, Start) = Out3
			GOSUB Assign.End
			RETURN
		    ELSE
			IF Strngs(Variable) = Nul THEN
			    IF LEN(Out3) = Start THEN
				Strngs(Variable) = Out3
				GOSUB Assign.End
				RETURN
			    END IF
			END IF
		    END IF
		END IF
	    END IF
	END IF
    END IF
END IF
RETURN

Assign.End:
IF LEN(Strng) THEN
    CALL Parse1(x#)
    Strng = "<extra token>"
    ERROR 92
END IF
RETURN

Define.Function:
Out2 = MID$(Out2, 6)
Out2 = LTRIM$(Out2)
Token.Index = 1
CALL Get.Token3
CALL Get.Token
CALL Parse1(Temp4#)
CALL Get.Token4
Func% = CINT(Temp4#)
IF Func% >= 1 AND Func% <= Max.FNs THEN
    Out2 = MID$(Out2, Token.Index)
    Out2 = LTRIM$(Out2)
    IF LEFT$(Out2, 1) = "=" THEN
	Out2 = MID$(Out2, 2)
	Definitions(Func%) = Out2
	Assign = True
	RETURN
    END IF
END IF
ERROR 145
END SUB

SUB Equate (Temp#)
Last.Token = False
Out3 = Nul ' reset string storage
Out4 = Nul ' reset string concatenate storage
Temp# = False ' reset result
Token.Index = 1 ' reset pointer to expression token
CALL Get.Token ' read next token
CALL Parse1(Temp#) ' entry to parse the expression
' parse leftover tokens
WHILE LEN(Strng)
    CALL Parse1(x#)
    ' toss away any unwanted tokens
    IF Last.Token = 3 THEN
	Strng = "<extra token>"
	ERROR 92
    END IF
WEND
END SUB

' remove statusline and exit.
SUB ExitProgram
ON LOCAL ERROR GOTO ExitError
' restore basic dta
InregsX.AX = &H1A00
InregsX.DS = BASIC.DTA.SEG
InregsX.DX = BASIC.DTA.OFF
CALL InterruptX(&H21, InregsX, OutregsX)
DEF SEG
' restore current drive/directory
CHDRIVE Current.Directory
CHDIR Current.Directory
ExitResume:
CALL ClearStatusLine
COLOR Plain, Black
CLS
CALL Save.Screen2(-1)
IF TempX = 0 THEN TempX = 1
LOCATE TempX, 1, 1
PRINT
COLOR White, Black
PRINT "Now exiting Whatis."
COLOR Plain, Black
END
ExitError:
 RESUME ExitResume
END SUB

SUB Filedate (V$)
 DIM ASCIIZ2 AS STRING * 260
 DIM ParseDTA AS DTAtype
 V$ = Nul
 IF LEN(Out3) THEN
    ASCIIZ2 = Out3 + CHR$(0)
    ' restore directory search dta
    InregsX.AX = &H1A00
    InregsX.DS = VARSEG(ParseDTA)
    InregsX.DX = VARPTR(ParseDTA)
    CALL InterruptX(&H21, InregsX, OutregsX)
    ' find first filename
    InregsX.AX = &H4E00
    InregsX.CX = &H37
    InregsX.DS = VARSEG(ASCIIZ2)
    InregsX.DX = VARPTR(ASCIIZ2)
    CALL InterruptX(&H21, InregsX, OutregsX)
    ' check findfirst error
    IF (OutregsX.Flags AND &H1) = &H0 THEN
       ' store file date
       Work! = ASC(MID$(ParseDTA.Filedate, 2, 1))
       Work! = Work! * &H100 + ASC(MID$(ParseDTA.Filedate, 1, 1))
       YearTemp! = INT(Work! / 512)
       MonthTemp! = INT((Work! AND &H1E0) / 32)
       DayTemp! = INT(Work! AND &H1F)
       YearTemp! = YearTemp! + 1980
       V$ = RIGHT$(STR$(MonthTemp! + 100), 2) + "-" + RIGHT$(STR$(DayTemp! + 100), 2) + "-" + MID$(STR$(YearTemp!), 2)
    END IF
    ' restore basic dta
    InregsX.AX = &H1A00
    InregsX.DS = BASIC.DTA.SEG
    InregsX.DX = BASIC.DTA.OFF
    CALL InterruptX(&H21, InregsX, OutregsX)
 END IF
END SUB

SUB Filetime (V$)
 DIM ASCIIZ2 AS STRING * 260
 DIM ParseDTA AS DTAtype
 V$ = Nul
 IF LEN(Out3) THEN
    ASCIIZ2 = Out3 + CHR$(0)
    ' restore directory search dta
    InregsX.AX = &H1A00
    InregsX.DS = VARSEG(ParseDTA)
    InregsX.DX = VARPTR(ParseDTA)
    CALL InterruptX(&H21, InregsX, OutregsX)
    ' find first filename
    InregsX.AX = &H4E00
    InregsX.CX = &H37
    InregsX.DS = VARSEG(ASCIIZ2)
    InregsX.DX = VARPTR(ASCIIZ2)
    CALL InterruptX(&H21, InregsX, OutregsX)
    ' check findfirst error
    IF (OutregsX.Flags AND &H1) = &H0 THEN
       Work! = ASC(MID$(ParseDTA.Filetime, 2, 1))
       Work! = Work! * &H100 + ASC(MID$(ParseDTA.Filetime, 1, 1))
       HourTemp! = INT(Work! / 2048)
       MinuteTemp! = INT((Work! AND &H7E0) / 32)
       SecondsTemp! = INT((Work! AND &H1F) / 2)
       V$ = RIGHT$(STR$(HourTemp! + 100), 2) + ":" + RIGHT$(STR$(MinuteTemp! + 100), 2) + ":" + RIGHT$(STR$(SecondsTemp! + 100), 2)
    END IF
    ' restore basic dta
    InregsX.AX = &H1A00
    InregsX.DS = BASIC.DTA.SEG
    InregsX.DX = BASIC.DTA.OFF
    CALL InterruptX(&H21, InregsX, OutregsX)
 END IF
END SUB

SUB FormatForm (T$, T#)
 ' 1.#INF and -1.#INF and -1.#IND and 1#QNAN
 IF INSTR(STR$(T#), "#") THEN
    T$ = LTRIM$(STR$(T#))
    EXIT SUB
 END IF
 ' store output value
 T$ = STR$(T#): x# = 0#
 ' check and count decimal values
 IF INSTR(T$, ".") THEN
    T$ = MID$(T$, INSTR(T$, ".") + 1)
    x# = LEN(T$)
 END IF
 IF x# = 0# THEN ' display with no decimal
    T$ = FORMAT$(T#, "#,##0")
 ELSE
    ' check and trim decimals
    Z$ = "#,##0." + STRING$(x#, "0")
    T$ = FORMAT$(T#, Z$)
    C$ = MID$(T$, INSTR(T$, ".") + 1)
    IF C$ = STRING$(LEN(C$), "0") THEN
       T$ = LEFT$(T$, INSTR(T$, ".") + 1)
    END IF
 END IF
END SUB

SUB Get.Bits (Temp#)
 DIM ASCIIZ2 AS STRING * 260
 DIM ParseDTA AS DTAtype
 ' store directory search dta
 InregsX.AX = &H1A00
 InregsX.DS = VARSEG(ParseDTA)
 InregsX.DX = VARPTR(ParseDTA)
 CALL InterruptX(&H21, InregsX, OutregsX)
 ' find first filename
 ASCIIZ2 = Out3 + CHR$(0)
 InregsX.AX = &H4E00
 InregsX.CX = &H37
 InregsX.DS = VARSEG(ASCIIZ2)
 InregsX.DX = VARPTR(ASCIIZ2)
 CALL InterruptX(&H21, InregsX, OutregsX)
 ' check findfirst error
 Temp# = 0#
 IF (OutregsX.Flags AND &H1) = &H0 THEN
    ' store file attributes
    Temp# = ASC(ParseDTA.FileAttr)
 END IF
 ' restore basic dta
 InregsX.AX = &H1A00
 InregsX.DS = BASIC.DTA.SEG
 InregsX.DX = BASIC.DTA.OFF
 CALL InterruptX(&H21, InregsX, OutregsX)
END SUB

' store default server name
SUB Get.Machine.Name (VarX$)
 VarX$ = Nul
 Default.Net = Nul
 InregsX.AX = &H5E00
 InregsX.DS = VARSEG(Default.Net)
 InregsX.DX = VARPTR(Default.Net)
 CALL InterruptX(&H21, InregsX, OutregsX)
 ' check error flag.
 IF (OutregsX.Flags AND &H1) = &H0 THEN
    ' read CH
    VarS = (OutregsX.CX AND &HFF00) / 256
    IF VarS > 0 THEN ' valid name.
       VarX$ = Default.Net
       Imbedded = INSTR(VarX$, CHR$(0))
       IF Imbedded THEN
	  VarX$ = LEFT$(VarX$, Imbedded - 1)
       END IF
       VarX$ = UCASE$(RTRIM$(VarX$))
       EXIT SUB
    END IF
 END IF
 ' check net path.
 IF VarX$ = Nul THEN
    Var$ = ENVIRON$("COMPUTERNAME")
    IF LEN(Var$) THEN
       VarX$ = Var$
    END IF
 END IF
END SUB

' returns next token in string form,
' increments pointer to next token,
' determines token type.
SUB Get.Token
Strng = Nul ' reset token
Token = False ' reset token type
' locate symbols
CALL Get.Token2(Token.Exists)
IF Token.Exists THEN
    EXIT SUB
END IF
IF Token.Index > LEN(Out2) THEN
    Strng = Nul
    Token = False
    Token.Index = Token.Index + 1
    EXIT SUB
END IF
' locate expression symbol
Token.Element$ = MID$(Out2, Token.Index, 1)
IF INSTR(Token.List, Token.Element$) THEN
    Token = 1 ' store token type
    Strng = Token.Element$ ' store token
    IF Strng = ";" THEN Strng = ","' force semicolon
    Token.Index = Token.Index + 1 ' increment pointer
    EXIT SUB
END IF
' locate expression is number
Token.Element$ = MID$(Out2, Token.Index, 1)
IF Token.Element$ >= "0" AND Token.Element$ <= "9" THEN
    ' increment token until token is other than number
    DO
	IF LEN(Token.Element$) = False THEN
	    EXIT DO
	END IF
	IF INSTR(Token.List, Token.Element$) THEN
	    EXIT DO
	END IF
	Strng = Strng + Token.Element$
	Token.Index = Token.Index + 1
	Token.Element$ = MID$(Out2, Token.Index, 1)
    LOOP
    Token = 2 ' store token type
    EXIT SUB
END IF
' locate expression is number beginning with decimal
Token.Element$ = MID$(Out2, Token.Index, 1)
IF Token.Element$ = "." THEN
    ' increment token until token is other than number
    DO
	IF LEN(Token.Element$) = False THEN
	    EXIT DO
	END IF
	IF INSTR(Token.List, Token.Element$) THEN
	    EXIT DO
	END IF
	Strng = Strng + Token.Element$
	Token.Index = Token.Index + 1
	Token.Element$ = MID$(Out2, Token.Index, 1)
    LOOP
    Token = 2 ' store token type
    EXIT SUB
END IF
' locate expression is alphabetic
Token.Element$ = UCASE$(MID$(Out2, Token.Index, 1))
IF Token.Element$ >= "A" AND Token.Element$ <= "Z" THEN
    ' increment token until token is other than alphabetic
    DO
	IF LEN(Token.Element$) = False THEN
	    EXIT DO
	END IF
	IF INSTR(Token.List, Token.Element$) THEN
	    EXIT DO
	END IF
	Strng = Strng + Token.Element$
	Token.Index = Token.Index + 1
	Token.Element$ = UCASE$(MID$(Out2, Token.Index, 1))
    LOOP
    Token = 3 ' store token type
    EXIT SUB
END IF
Token = False
Strng = Token.Element$
Token.Index = Token.Index + 1
END SUB

' increments pointer to next token.
SUB Get.Token2 (Token.Exists)
Token.Exists = False
Stored.Token$ = Out2

' gets three-symbol tokens,

' locate expression symbols
Next.Token$ = MID$(Stored.Token$, Token.Index, 3)
SELECT CASE Next.Token$
    CASE "^-="
	Token = 1 ' store token type
	Strng = "^-=" ' store token
	Token.Index = Token.Index + 3 ' increment pointer
	Token.Exists = True
	EXIT SUB
    CASE "^+="
	Token = 1 ' store token type
	Strng = "^+=" ' store token
	Token.Index = Token.Index + 3 ' increment pointer
	Token.Exists = True
	EXIT SUB
END SELECT

' gets two-symbol tokens,

' locate expression symbols
Next.Token$ = MID$(Stored.Token$, Token.Index, 2)
SELECT CASE Next.Token$
    CASE "<<"
	Strng = "<<" ' store token
	GOTO End.Token2
    CASE ">>"
	Strng = ">>" ' store token
	GOTO End.Token2
    CASE "--"
	Strng = "--" ' store token
	GOTO End.Token2
    CASE "++"
	Strng = "++" ' store token
	GOTO End.Token2
    CASE "**"
	Strng = "**" ' store token
	GOTO End.Token2
    CASE "//"
	Strng = "//" ' store token
	GOTO End.Token2
    CASE ">=", "=>"
	Strng = ">=" ' store token
	GOTO End.Token2
    CASE "<=", "=<"
	Strng = "<=" ' store token
	GOTO End.Token2
    CASE "<>", "><"
	Strng = "<>" ' store token
	GOTO End.Token2
    CASE "=="
	Strng = "==" ' store token
	GOTO End.Token2
    CASE "||"
	Strng = "||" ' store token
	GOTO End.Token2
    CASE "##"
	Strng = "##" ' store token
	GOTO End.Token2
    CASE "^^"
	Strng = "^^" ' store token
	GOTO End.Token2
    CASE "^="
	Strng = "^=" ' store token
	GOTO End.Token2
END SELECT
EXIT SUB
End.Token2:
 Token = 1 ' store token type
 Token.Index = Token.Index + 2 ' increment pointer
 Token.Exists = True
END SUB

' verifies next token is opening parenthesis.
SUB Get.Token3
Strng = Nul
Token = False
IF Token.Index > LEN(Out2) THEN
    Strng = "<missing opening parenthesis>"
    ERROR 92
END IF
' locate parenthesis symbol
Token.Element$ = MID$(Out2, Token.Index, 1)
IF Token.Element$ = "(" THEN
    Token = 1 ' store token type
    Strng = Token.Element$ ' store token
    Token.Index = Token.Index + 1 ' increment pointer
    EXIT SUB
END IF
Strng = "<missing opening parenthesis>"
ERROR 92
END SUB

' verifies last token is closing parenthesis.
SUB Get.Token4
' locate parenthesis symbol
IF Strng = ")" THEN
    EXIT SUB
END IF
Strng = "<missing closing token>"
ERROR 92
END SUB

SUB GetFilesize (Temp#)
 DIM ASCIIZ2 AS STRING * 260
 DIM ParseDTA AS DTAtype
 Temp# = DFalse
 IF LEN(Out3) THEN
    ASCIIZ2 = Out3 + CHR$(0)
    ' restore directory search dta
    InregsX.AX = &H1A00
    InregsX.DS = VARSEG(ParseDTA)
    InregsX.DX = VARPTR(ParseDTA)
    CALL InterruptX(&H21, InregsX, OutregsX)
    ' find first filename
    InregsX.AX = &H4E00
    InregsX.CX = &H27
    InregsX.DS = VARSEG(ASCIIZ2)
    InregsX.DX = VARPTR(ASCIIZ2)
    CALL InterruptX(&H21, InregsX, OutregsX)
    ' check findfirst error
    IF (OutregsX.Flags AND &H1) = &H0 THEN
       ' store file size
       Temp# = ASC(MID$(ParseDTA.FileSize, 4, 1))
       Temp# = Temp# * &H100 + ASC(MID$(ParseDTA.FileSize, 3, 1))
       Temp# = Temp# * &H100 + ASC(MID$(ParseDTA.FileSize, 2, 1))
       Temp# = Temp# * &H100 + ASC(MID$(ParseDTA.FileSize, 1, 1))
    END IF
    ' restore basic dta
    InregsX.AX = &H1A00
    InregsX.DS = BASIC.DTA.SEG
    InregsX.DX = BASIC.DTA.OFF
    CALL InterruptX(&H21, InregsX, OutregsX)
 END IF
END SUB

' function to match substring with ?, * characters in substring.
SUB InstrSUB (Var, Var1$, Var2$)
' store string match variables
S2$ = LCASE$(Var2$)
S3$ = LCASE$(Var1$)

' check default instr
IF INSTR(S2$, "*") = 0 THEN
   IF INSTR(S2$, "?") = 0 THEN
      Var = INSTR(S3$, S2$)
      EXIT SUB
   END IF
END IF

Var = -1 ' assume match

' see if S2$ matches in S3$ with substrings
FOR S3 = 1 TO LEN(S3$)
   S1$ = MID$(S3$, S3)
   P1 = 1 ' pointer to S1$
   P2 = 1 ' pointer to S2$
   P3 = S3 ' position match
   DO
      ' check token beyond string
      IF P2 > LEN(S2$) THEN
	 Var = P3
	 EXIT SUB
      END IF

      ' check character in S2$ at P2
      V$ = MID$(S2$, P2, 1)
      SELECT CASE V$
      CASE "*" ' global character
	 ' scan to next char
	 IF P2 > LEN(S2$) THEN
	    EXIT DO
	 END IF
	 S4$ = MID$(S2$, P2 + 1, 1)
	 SELECT CASE S4$
	 CASE "*", "?"
	    P2 = P2 + 1
	 CASE ELSE
	    DO
	       IF MID$(S1$, P1, 1) = S4$ THEN
		  EXIT DO
	       END IF
	       IF P1 >= LEN(S1$) THEN
		  EXIT DO
	       END IF
	       P1 = P1 + 1
	    LOOP
	    P3 = P1
	    P2 = P2 + 1
	 END SELECT
      CASE "?" ' wildcard character
	 P1 = P1 + 1
	 P2 = P2 + 1
      CASE ELSE ' ascii character
	 IF MID$(S1$, P1, 1) <> V$ THEN ' no match
	    EXIT DO
	 END IF
	 P1 = P1 + 1
	 P2 = P2 + 1
      END SELECT
   LOOP
NEXT
Var = 0 ' no match
END SUB

' reads character directly from keyboard including extended keys.
FUNCTION KeyboardChar$
 IF KeyIS = 0 THEN
    KeyboardChar$ = Nul
    EXIT FUNCTION
 END IF
 InregsX.AX = &H0
 CALL InterruptX(&H16, InregsX, OutregsX)
 Key1 = (OutregsX.AX AND &HFF)
 IF Key1 = False THEN ' extended key
    Key2 = (OutregsX.AX AND &HFF00) / 256
    KeyboardChar$ = CHR$(0) + CHR$(Key2)
 ELSE
    KeyboardChar$ = CHR$(Key1)
    IF Key1 = 3 THEN
       ControlBreak = True
    END IF
 END IF
 Flag = BreakIS ' check control-break
END FUNCTION

FUNCTION KeyboardLine$ (VarZ%)
REM uses direct keyboard input to get an input line
REM   and edit multi-dimensional command history arrays with line-wrapping,
REM     where Temp0 (history array type) equals:
REM       1=<n/a>
REM       2=<n/a>
REM       3=whatis prompt

REM  VarZ% equals True to store command in history.

History.Count2(Temp0) = 0 ' reset
Var$ = Nul
StartTime2# = Timeit!
DO
    RunTime2# = Timeit! - StartTime2#
    IF RunTime2# < 0# THEN RunTime2# = RunTime2# + 86400#
    IF RunTime2# >= 1# THEN
       StartTime2# = Timeit!
       Var1 = CSRLIN
       Var2 = POS(0)
       CALL DisplayStatusLine(Var1, Var2)
    END IF
    Var2$ = KeyboardChar$ ' get character
    SELECT CASE LEN(Var2$)
	CASE 0
	    Eat$ = Nul
	CASE 1
	    SELECT CASE Var2$
		CASE CHR$(3), CHR$(11) ' control-c/control-k
		    GOSUB Erase.Line
		    COLOR Red, Black
		    PRINT "*break*"
		    SOUND 900, 6
		    COLOR Yellow, Black
		    LOCATE , 1, 1
		    PRINT ">";
		    LOCATE , , 1
		CASE CHR$(9) ' tab
		    IF LEN(Var$) + 8 <= 256 THEN
			FOR VarX = 1 TO 8
			    PRINT " ";
			    LOCATE , , 1
			    Var$ = Var$ + " "
			NEXT
		    END IF
		CASE CHR$(13) ' return
		    IF VarZ% THEN
		       GOSUB Enter.History
		    END IF
		    EXIT DO
		CASE CHR$(8) ' backspace
		    IF LEN(Var$) > 0 THEN
			Var$ = LEFT$(Var$, LEN(Var$) - 1)
			GOSUB Back.Space
		    END IF
		CASE CHR$(27) ' escape
		    GOSUB Erase.Line
		CASE ELSE ' store input
		    IF LEN(Var$) < 256 THEN
			PRINT Var2$;
			LOCATE , , 1
			Var$ = Var$ + Var2$
		    END IF
	    END SELECT
	CASE 2
	    SELECT CASE ASC(RIGHT$(Var2$, 1))
		CASE 147 ' control-delete
		   CALL ExitProgram
		   END
		CASE 83 ' delete
		   Debug = NOT Debug
		   Var1 = CSRLIN
		   Var2 = POS(0)
		   CALL DisplayStatusLine(Var1, Var2)
		CASE 76, 82 ' keypad-5, insert
		   Var1 = CSRLIN
		   Var2 = POS(0)
		   Ascii$ = Nul
		   GOSUB Display.Ascii.Line
		   DO
		      Var3$ = KeyboardChar$
		      IF Var3$ = CHR$(13) THEN ' return
			 EXIT DO
		      END IF
		      IF Var3$ = CHR$(27) THEN ' escape
			 Var3$ = Nul
			 IF LEN(Ascii$) THEN
			    Ascii$ = Nul
			    GOSUB Display.Ascii.Line
			 END IF
		      END IF
		      IF Var3$ = CHR$(8) THEN ' backspace
			 Var3$ = Nul
			 IF LEN(Ascii$) THEN
			    Ascii$ = LEFT$(Ascii$, LEN(Ascii$) - 1)
			    GOSUB Display.Ascii.Line
			 END IF
		      END IF
		      IF LCASE$(Var3$) = "x" OR LCASE$(Var3$) = "o" OR LCASE$(Var3$) = "b" THEN
			 IF LEN(Ascii$) = 0 THEN
			    Ascii$ = Ascii$ + LCASE$(Var3$)
			    GOSUB Display.Ascii.Line
			 END IF
		      END IF
		      IF LCASE$(Var3$) >= "a" AND LCASE$(Var3$) <= "f" THEN
			 IF LEN(Ascii$) >= 1 AND LEN(Ascii$) < 3 THEN
			    IF LEFT$(Ascii$, 1) = "x" THEN
			       Ascii$ = Ascii$ + Var3$
			       GOSUB Display.Ascii.Line
			    END IF
			 END IF
		      END IF
		      IF LEFT$(Ascii$, 1) = "o" THEN
			 IF Var3$ >= "0" AND Var3$ <= "7" THEN
			    IF LEN(Ascii$) < 4 THEN
			       Ascii$ = Ascii$ + Var3$
			       GOSUB Display.Ascii.Line
			    END IF
			 END IF
		      ELSE
			 IF LEFT$(Ascii$, 1) = "b" THEN
			    IF Var3$ = "0" OR Var3$ = "1" THEN
			       IF LEN(Ascii$) < 9 THEN
				  Ascii$ = Ascii$ + Var3$
				  GOSUB Display.Ascii.Line
			       END IF
			    END IF
			 ELSE
			    IF Var3$ >= "0" AND Var3$ <= "9" THEN
			       IF LEN(Ascii$) < 3 THEN
				  Ascii$ = Ascii$ + Var3$
				  GOSUB Display.Ascii.Line
			       END IF
			    END IF
			 END IF
		      END IF
		   LOOP
		   CALL DisplayStatusLine(Var1, Var2)

		   ' insert ascii character
		   IF LEFT$(Ascii$, 1) = "x" THEN ' hexidecimal
		      Ascii.Value = INT(VAL("&H" + MID$(Ascii$, 2)))
		   ELSE
		      IF LEFT$(Ascii$, 1) = "o" THEN ' octal
			 Ascii.Value = INT(VAL("&O" + MID$(Ascii$, 2)))
		      ELSE
			 IF LEFT$(Ascii$, 1) = "b" THEN ' binary
			    Ascii$ = MID$(Ascii$, 2)
			    GOSUB Binary.String
			 ELSE
			    Ascii.Value = INT(VAL(Ascii$))
			 END IF
		      END IF
		   END IF
		   IF Ascii.Value >= 0 AND Ascii.Value <= 255 THEN
		      IF Ascii.Value = 10 OR Ascii.Value = 13 THEN
			 Eat$ = Nul
		      ELSE
			 IF LEN(Var$) < 256 THEN
			    PRINT CHR$(Ascii.Value);
			    LOCATE , , 1
			    Var$ = Var$ + CHR$(Ascii.Value)
			 END IF
		      END IF
		   END IF
		CASE 77 ' right
		    GOSUB Get.Count
		    IF Var1 = 0 THEN
		       PRINT "Text not found."
		       PRINT ">";
		    END IF
                    ' return blank line at end of history
		    IF Var1 > 0 THEN
                       If Var1 = History.Count2(Temp0) Then
                          EndOfHistoryFlag(Temp0) = -1
                          GoSub Erase.Line
                          Var$ = ""
                          Var1 = 0
                          Gosub PrintLine
                       End If
                    End If
                    ' move to next history
		    IF Var1 > 0 THEN
			SELECT CASE History.Count2(Temp0)
			    CASE 0
				GOSUB Erase.Line
				History.Count2(Temp0) = Var1
				Var$ = RTRIM$(History2(Temp0, Var1))
				GOSUB PrintLine
			    CASE 1 TO Max.History - 1
				IF History.Count2(Temp0) + 1 <= Var1 THEN
				    GOSUB Erase.Line
				    History.Count2(Temp0) = History.Count2(Temp0) + 1
				    Var$ = RTRIM$(History2(Temp0, History.Count2(Temp0)))
				    GOSUB PrintLine
				ELSE
				    IF History.Count2(Temp0) = Var1 THEN
					GOSUB Erase.Line
					Var$ = RTRIM$(History2(Temp0, Var1))
					GOSUB PrintLine
				    END IF
				END IF
			    CASE Max.History
				GOSUB Erase.Line
				Var$ = RTRIM$(History2(Temp0, Max.History))
				GOSUB PrintLine
			END SELECT
		    END IF
		CASE 75 ' left
		    GOSUB Get.Count
		    IF Var1 = 0 THEN
		       PRINT "Text not found."
		       PRINT ">";
		    END IF
                    If Var1 > 0 Then
                       If EndOfHistoryFlag(Temp0) Then
                          EndOfHistoryFlag(Temp0) = 0
                          GoSub Erase.Line
                          History.Count2(Temp0) = Var1
                          Var$ = RTrim$(History2(Temp0, Var1))
                          GOSUB PrintLine
                          Var1 = 0
                       End If
                    End If
		    IF Var1 > 0 THEN
			SELECT CASE History.Count2(Temp0)
			    CASE 0
				GOSUB Erase.Line
				History.Count2(Temp0) = Var1
				Var$ = RTRIM$(History2(Temp0, Var1))
				GOSUB PrintLine
			    CASE 1
				GOSUB Erase.Line
				Var$ = RTRIM$(History2(Temp0, 1))
				GOSUB PrintLine
			    CASE ELSE
				IF History.Count2(Temp0) - 1 > 0 THEN
				    GOSUB Erase.Line
				    History.Count2(Temp0) = History.Count2(Temp0) - 1
				    Var$ = RTRIM$(History2(Temp0, History.Count2(Temp0)))
				    GOSUB PrintLine
				END IF
			END SELECT
		    END IF
		CASE 73 ' pageup
		    GOSUB Get.Count
		    IF Var1 = 0 THEN
		       PRINT "Text not found."
		       PRINT ">";
		    END IF
		    IF Var1 > 0 THEN
		       PRINT
		       COLOR White
		       Var2$ = "Enter search text(? and * allowed):"
		       PRINT Var2$;
		       LINE INPUT Ascii$
		       COLOR Yellow
		       IF LEN(Ascii$) THEN
			  FOR V = 1 TO Var1
			     V1$ = LCASE$(History2(Temp0, V))
			     V2$ = LCASE$(Ascii$)
			     CALL InstrSUB(Var, V1$, V2$)
			     IF Var THEN
				Var$ = RTRIM$(History2(Temp0, V))
				KeyboardLine$ = Var$
				PRINT ">";
				GOSUB PrintLine
				EXIT FUNCTION
			     END IF
			  NEXT
		       END IF
		       PRINT "Text not found."
		       PRINT ">";
		   END IF
		CASE 81 ' pagedown
		    GOSUB Get.Count
		    IF Var1 = 0 THEN
		       PRINT "Text not found."
		       PRINT ">";
		    END IF
		    IF Var1 > 0 THEN
		       PRINT
		       COLOR White
		       Var2$ = "Enter search text(? and * allowed):"
		       PRINT Var2$;
		       LINE INPUT Ascii$
		       COLOR Yellow
		       IF LEN(Ascii$) THEN
			  FOR V = Var1 TO 1 STEP -1
			     V1$ = LCASE$(History2(Temp0, V))
			     V2$ = LCASE$(Ascii$)
			     CALL InstrSUB(Var, V1$, V2$)
			     IF Var THEN
				Var$ = RTRIM$(History2(Temp0, V))
				KeyboardLine$ = Var$
				PRINT ">";
				GOSUB PrintLine
				EXIT FUNCTION
			     END IF
			  NEXT
		       END IF
		       PRINT "Text not found."
		       PRINT ">";
		   END IF
		CASE 71 ' home
		    GOSUB Get.Count
		    IF Var1 = 0 THEN
		       PRINT "Text not found."
		       PRINT ">";
		    END IF
		    IF Var1 > 0 THEN
			GOSUB Erase.Line
			History.Count2(Temp0) = 1
			Var$ = RTRIM$(History2(Temp0, 1))
			GOSUB PrintLine
		    END IF
		CASE 79 ' end
		    GOSUB Get.Count
		    IF Var1 = 0 THEN
		       PRINT "Text not found."
		       PRINT ">";
		    END IF
		    IF Var1 > 0 THEN
			GOSUB Erase.Line
			History.Count2(Temp0) = Var1
			Var$ = RTRIM$(History2(Temp0, Var1))
			GOSUB PrintLine
		    END IF
	    END SELECT
    END SELECT
    LOCATE , , 1
LOOP

' return input line
KeyboardLine$ = Var$
EXIT FUNCTION

Binary.String:
  Ascii.Value = False
  Binary.Power = False
  FOR Binary.Digit = LEN(Ascii$) TO 1 STEP -1
      OutX$ = MID$(Ascii$, Binary.Digit, 1)
      SELECT CASE OutX$
	  CASE "0"
	      ' nul
	  CASE "1"
	      Ascii.Value = Ascii.Value + 2 ^ Binary.Power
      END SELECT
      Binary.Power = Binary.Power + 1
  NEXT
RETURN

Display.Ascii.Line:
  COLOR White, Blue
  LOCATE Max.Row, 1, 1
  IF LEFT$(Ascii$, 1) = "x" THEN
     Var2$ = "Enter 2-digit ascii code: " + Ascii$
  ELSE
     IF LEFT$(Ascii$, 1) = "o" THEN
	Var2$ = "Enter 3-digit ascii code: " + Ascii$
     ELSE
	IF LEFT$(Ascii$, 1) = "b" THEN
	   Var2$ = "Enter 8-digit ascii code: " + Ascii$
	ELSE
	   Var2$ = "Enter 3-digit ascii code: " + Ascii$
	END IF
     END IF
  END IF
  VarC = LEN(Var2$) + 1
  Var2$ = Var2$ + SPACE$(80 - LEN(Var2$))
  PRINT Var2$;
  LOCATE Max.Row, VarC, 1
  RETURN

PrintLine:
 FOR VarX = 1 TO LEN(Var$)
    PRINT MID$(Var$, VarX, 1);
 NEXT
 RETURN

Enter.History:
IF LEN(Var$) THEN
    GOSUB Get.Count
    ' init
    EndOfHistoryFlag(Temp0) = -1
    IF Var1 = 0 THEN
	History.Count2(Temp0) = 1
	History2(Temp0, 1) = Var$
    ELSE
	' add history
	IF History.Count2(Temp0) = 0 THEN
	    IF Var1 < Max.History THEN
		History.Count2(Temp0) = Var1 + 1
		History2(Temp0, Var1 + 1) = Var$
	    ELSE
		FOR Var = 1 TO Max.History - 1
		    History2(Temp0, Var) = History2(Temp0, Var + 1)
		NEXT
		History.Count2(Temp0) = Max.History
		History2(Temp0, Max.History) = Var$
	    END IF
	ELSE
	    ' pack history
	    IF Var1 = Max.History THEN
		FOR Var = History.Count2(Temp0) TO Max.History - 1
		    History2(Temp0, Var) = History2(Temp0, Var + 1)
		NEXT
		History.Count2(Temp0) = Max.History
		History2(Temp0, Max.History) = Var$
	    ELSE
		FOR Var = History.Count2(Temp0) TO Var1 - 1
		    History2(Temp0, Var) = History2(Temp0, Var + 1)
		NEXT
		History2(Temp0, Var1) = Var$
	    END IF
	END IF
    END IF
END IF
RETURN

Erase.Line:
FOR VarX = 1 TO LEN(Var$)
    GOSUB Back.Space
NEXT
Var$ = Nul
RETURN

Back.Space:
IF POS(0) > 1 THEN ' check cursor location
    LOCATE CSRLIN, POS(0) - 1, 0
    PRINT " ";
    LOCATE CSRLIN, POS(0) - 1, 1
ELSE ' line-wrap
    LOCATE CSRLIN - 1, 80, 0
    PRINT " ";
    LOCATE CSRLIN - 1, 80, 1
END IF
RETURN

Get.Count:
Var1 = 0
FOR Var2 = Max.History TO 1 STEP -1
    IF RTRIM$(History2(Temp0, Var2)) <> Nul THEN
	Var1 = Var2
	EXIT FOR
    END IF
NEXT
RETURN
END FUNCTION

' checks keyboard buffer
FUNCTION KeyIS
 InregsX.AX = &H100
 CALL InterruptX(&H16, InregsX, OutregsX)
 IF (OutregsX.Flags AND &H40) = &H40 THEN
    KeyIS = False
 ELSE
    KeyIS = True
 END IF
END FUNCTION

SUB MD.Directory
 CALL Get.Token3
 CALL Get.Token
 CALL Parse1(Temp#)
 CALL Get.Token4
 Hidden = True
 MKDIR Out3
 Last.Token = False
END SUB

' prompts for multiple input characters
SUB More.Prompt (Input.String$, Input.Mask$, Output.String$)
 PRINT Input.String$; " ";
 Input.Char$ = Nul
 StartTime2# = Timeit!
 DO
    RunTime2# = Timeit! - StartTime2#
    IF RunTime2# < 0# THEN RunTime2# = RunTime2# + 86400#
    IF RunTime2# >= 1# THEN
       StartTime2# = Timeit!
       Var1 = CSRLIN
       Var2 = POS(0)
       CALL DisplayStatusLine(Var1, Var2)
    END IF
    LOCATE , , 1
    Input.Char$ = KeyboardChar$
    IF LEN(Input.Char$) THEN
       IF INSTR(Input.Mask$, Input.Char$) THEN
	  IF Input.Char$ = CHR$(13) THEN
	     PRINT
	  ELSE
	     PRINT Input.Char$
	  END IF
	  Output.String$ = Input.Char$
	  EXIT DO
       END IF
    END IF
 LOOP
END SUB

SUB Name.Expression
 CALL Get.Token3
 CALL Get.Token
 CALL Parse1(Temp3#)
 IF Last.Token = False THEN
    Var4$ = Out3
    IF Strng = "," THEN
       Last.Token = False
       CALL Get.Token
       CALL Parse1(Temp4#)
       IF Last.Token = False THEN
	  CALL Get.Token4
	  Out4 = Var4$
	  IF Out3 = Nul THEN
	     ERROR 76
	     EXIT SUB
	  END IF
	  IF Out4 = Nul THEN
	     ERROR 76
	     EXIT SUB
	  END IF
	  Hidden = -1
	  NAME Out4 AS Out3
	  EXIT SUB
       END IF
    END IF
 END IF
 ERROR 76
END SUB

' simple subroutine to display a file
SUB Out.File (Filename$, V)
 ON LOCAL ERROR GOTO Open.Error
 IF V THEN Z = Max.Row - 2
 IF V = 0 THEN Z = 19
 CLOSE #1
 OPEN Filename$ FOR INPUT AS #1
 DO UNTIL EOF(1)
    LINE INPUT #1, Var$
    x = x + 1
    COLOR White, Black
    PRINT Var$
    IF x = Z THEN
       x = 0
       IF C = 0 THEN
	  COLOR Yellow, Black
	  PRINT MorePrompt3;
	  DO
	     x$ = INKEY$
             Select Case Ucase$(x$)
             Case "C"
                PRINT : C = -1: EXIT DO
             Case "Q"
                PRINT : Close #Q : EXIT SUB
             Case "N"
                PRINT : Close #Q : EXIT SUB
             Case "Y"
                PRINT : EXIT DO
             Case " "
                PRINT : EXIT DO
             Case Chr$(13)
                PRINT : EXIT DO
             Case "?"
                Call MorePromptHelp
             End Select
	  LOOP
       END IF
    END IF
 LOOP
 EXIT SUB
Open.Error:
 PRINT Filename$; " not found."
 EXIT SUB
END SUB

Sub MorePromptHelp
 PRINT
 COLOR 15
 PRINT " Y = Yes"
 PRINT " N = No"
 PRINT " Q = Quit"
 PRINT " C = Continuous"
 PRINT "  <space> = continue"
 PRINT "  <enter> = continue"
 COLOR 14
 PRINT MorePrompt3;
End Sub

SUB Output.Port
 CALL Get.Token3
 CALL Get.Token
 CALL Parse1(Temp#)
 Port.Address = CINT(Temp#)
 CALL Get.Token
 CALL Parse1(Temp#)
 Port.Byte = CINT(Temp#)
 OUT Port.Address, Port.Byte
 CALL Get.Token4
 Last.Token = True
END SUB

SUB Wait.Port
 CALL Get.Token3
 CALL Get.Token
 CALL Parse1(Temp#)
 Port.Address = CINT(Temp#)
 CALL Get.Token
 CALL Parse1(Temp#)
 And.Value = CINT(Temp#)
 IF Strng = "," THEN
    CALL Get.Token
    CALL Parse1(Temp#)
    Xor.Value = CINT(Temp#)
    WAIT Port.Address, And.Value, Xor.Value
 ELSE
    WAIT Port.Address, And.Value
 END IF
 CALL Get.Token4
 Last.Token = True
END SUB

' parses single letter tokens
SUB Parse.Alphabetic1 (Temp#)
SELECT CASE UCASE$(Strng) ' test variable symbol
    CASE CHR$(10), CHR$(13)
	CALL Get.Token
    CASE "A" TO "Z"
	Element = ASC(UCASE$(Strng)) - 64
	Element.Type$ = MID$(Out2, Token.Index, 1)
	SELECT CASE Element.Type$
	    CASE "("
		CALL Get.Token
		CALL Get.Token
		CALL Parse1(Temp#)
		IF Strng = ")" THEN
		    Temp# = Arrays(Element, CINT(Temp#))
		    Last.Token = True
		ELSE
		    Strng = "<missing closing token>"
		    ERROR 92
		END IF
	    CASE "["
		CALL Get.Token
		CALL Get.Token
		CALL Parse1(Temp#)
		IF Strng = "]" THEN
		    Temp# = Arrays(Element, CINT(Temp#))
		    Last.Token = True
		ELSE
		    Strng = "<missing closing token>"
		    ERROR 92
		END IF
	    CASE "{"
		CALL Get.Token
		CALL Get.Token
		CALL Parse1(Temp#)
		IF Strng = "}" THEN
		    Temp# = Arrays(Element, CINT(Temp#))
		    Last.Token = True
		ELSE
		    Strng = "<missing closing token>"
		    ERROR 92
		END IF
	    CASE ELSE
		Temp# = Variables(Element)
		Last.Token = True
	END SELECT
    CASE ELSE
	ERROR 92
END SELECT
END SUB

' parses two-letter tokens
SUB Parse.Alphabetic2 (Temp#)
SELECT CASE UCASE$(Strng)
    ' user defined functions
    CASE "FN"
	CALL ParseFN(Temp#)
    CASE "FZ"
	CALL ParseFZ(Temp#)
	' functions w/o parameters
    CASE "PI" ' calculate PI
	Temp# = PI#
	Last.Token = True
    CASE "EX" ' calculate E
	Temp# = EXP(1#)
	Last.Token = True
	' special case functions
    CASE "CD"
	CALL CD.Directory
    CASE "MD"
	CALL MD.Directory
    CASE "RD"
	CALL RD.Directory
    CASE "OR"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Number# = Temp#
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = Number# OR Temp#
	CALL Get.Token4
	Last.Token = True
    CASE ELSE
	IF MID$(Strng, 2, 1) = "$" THEN
	    SELECT CASE MID$(Out2, Token.Index, 1)
		CASE "(", "[", "{"
		    Strng = "<bad string token>"
		    ERROR 92
	    END SELECT
	    SELECT CASE UCASE$(LEFT$(Strng, 1))
		CASE "A" TO "Z"
		    Element = ASC(UCASE$(LEFT$(Strng, 1))) - 64
		    Out3 = Strngs(Element)
		    Last.Token = False
	    END SELECT
	ELSE
	    ERROR 92
	END IF
END SELECT
END SUB

' parses three and greater letter tokens
SUB Parse.Alphabetic3 (Temp#)
SELECT CASE UCASE$(Strng)
 ' special sound function
 CASE "WAV"
    CALL Get.Token3
    CALL Get.Token
    CALL Parse1(Temp#)
    IF Last.Token THEN ' numeric
       VarX = CINT(Temp#)
       IF VarX >= 1 AND VarX <= 12 THEN
	  CALL Playwav(VarX, "")
       ELSE
	  ERROR 5
       END IF
    ELSE
       ERROR 5
    END IF
    CALL Get.Token4
    Temp# = -1#
    Out3 = "Sound"
    Last.Token = False
 CASE "WAV$"
    CALL Get.Token3
    CALL Get.Token
    CALL Parse1(Temp#)
    IF Last.Token = 0 THEN ' string
       VarX$ = Out3
       CALL Playwav(0, VarX$)
    ELSE
       ERROR 5
    END IF
    CALL Get.Token4
    Temp# = -1#
    Out3 = "Sound"
    Last.Token = False

    ' sound functions
    CASE "PLAY"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	PLAY "X" + VARPTR$(Out3)
	CALL Get.Token4
	Out3 = "Sound"
	Last.Token = False
    CASE "SOUND"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	CALL Get.Token
	CALL Parse1(Temp2#)
	SOUND Temp#, Temp2#
	CALL Get.Token4
	Out3 = "Sound"
	Last.Token = False
    CASE "BEEP"
	SOUND 900, 6
	Out3 = "Beep"
	Last.Token = False
    CASE "INKEY$"
	Out3 = INKEY$
	Last.Token = False
    CASE "MELODY$"
	Out3 = "o3 L8 E D+ E D+ E o2 B o3 D C L2 o2 A"
	Last.Token = False
    ' constant functions
    CASE "AUTHOR$"
	Out3 = "Erik Jon Oredson"
	Last.Token = False
    CASE "EMAIL$"
	Out3 = "eoredson@gmail.com"
	Last.Token = False
    CASE "URL$"
	Out3 = "www.filegate.net"
	Last.Token = False
    CASE "MACHINE$"
	CALL Get.Machine.Name(VarX$)
	Out3 = VarX$
	Last.Token = False
    CASE "COMMAND$"
	Out3 = COMMAND$
	Last.Token = False
    CASE "COMMANDLINE$"
	CALL Read.Command(Command.Line$)
	Out3 = Command.Line$
	Last.Token = False
'    CASE "DSPVERSION"
'        CALL ReadVersion (DSPVersion!)
'        Out3 = STR$(DSPVersion!)
'        Last.Token = False
    CASE "QUOTE$"
	Out3 = Quote$
	Last.Token = False
    CASE "VERSION$"
	Out3 = Version$
	Last.Token = False
    CASE "RELEASE$"
	Out3 = Release$
	Last.Token = False
    CASE "PROGRAM$"
	Out3 = "WHATIS"
	Last.Token = False
    CASE "PUBLISH$"
        Out3 = "01/01/2023"
	Last.Token = False
    CASE "YEARRELEASE$"
        Out3 = "2024"
	Last.Token = False
    CASE "COOKIE$"
	IF RND > .5 THEN
	   Out3 = "Thank you for all the cookies!"
	ELSE
	   Out3 = "I like sugar cookies!"
	END IF
	Last.Token = False
    CASE "ERROR"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	CALL Get.Token4

	CALL Get.Token
	IF Strng <> Nul THEN
	   Strng = "<extra closing token>"
	   ERROR 92
	ELSE
	   Strng = "<error>"
	   ERROR CINT(Temp#)
	END IF
	Last.Token = True

	' functions w/o parameters
    CASE "NUL", "NULL"
	Out3 = Nul
	Last.Token = False
    CASE "END", "STOP", "QUIT"
       IF DataError THEN
	  CALL ExitProgram
	  END
       END IF
    CASE "FALSE"
	Temp# = FalseD
	Last.Token = True
    CASE "TRUE"
	Temp# = TrueD
	Last.Token = True
    CASE "MAXARRAYS"
	Temp# = Max.Arrays
	Last.Token = True
    CASE "MAXHISTORY"
	Temp# = Max.History
	Last.Token = True
    CASE "MAXRECURSE"
	Temp# = Max.Recurse
	Last.Token = True
    CASE "MAXFNS"
	Temp# = Max.FNs
	Last.Token = True
    CASE "MAXFNLENGTH"
	Temp# = Max.FN.Length
	Last.Token = True
    CASE "TIMER"
	Temp# = TIMER
	Last.Token = True
    CASE "STACK"
	Temp# = STACK
	Last.Token = True
    CASE "NOW"
	Temp# = Now
	Last.Token = True
    CASE "RND"
	Temp# = RND
	Last.Token = True
	' boolean functions
    CASE "AND"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Number# = Temp#
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = Number# AND Temp#
	CALL Get.Token4
	Last.Token = True
    CASE "MOD"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Number# = Temp#
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = Number# MOD Temp#
	CALL Get.Token4
	Last.Token = True
    CASE "NOR"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Number# = Temp#
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = NOT (Number# OR Temp#)
	CALL Get.Token4
	Last.Token = True
    CASE "NON"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Number# = Temp#
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = NOT (Number# IMP Temp#)
	CALL Get.Token4
	Last.Token = True
    CASE "XAN"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Number# = Temp#
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = NOT (Number# AND Temp#)
	CALL Get.Token4
	Last.Token = True
    CASE "XOR"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Number# = Temp#
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = Number# XOR Temp#
	CALL Get.Token4
	Last.Token = True
    CASE "IMP"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Number# = Temp#
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = Number# IMP Temp#
	CALL Get.Token4
	Last.Token = True
    CASE "EQV"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Number# = Temp#
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = Number# EQV Temp#
	CALL Get.Token4
	Last.Token = True
    CASE "NOT"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = NOT Temp#
	CALL Get.Token4
	Last.Token = True
	' third-order boolean
    CASE "TAND"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Number1# = Temp#
	CALL Get.Token
	CALL Parse1(Temp#)
	Number2# = Temp#
	CALL Get.Token4
	SELECT CASE Number1#
	    CASE -1, 0, 1
		SELECT CASE Number2#
		    CASE -1, 0, 1
			Temp# = Number1# AND Number2#
			Last.Token = True
			EXIT SUB
		END SELECT
	END SELECT
	ERROR 5
    CASE "TOR"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Number1# = Temp#
	CALL Get.Token
	CALL Parse1(Temp#)
	Number2# = Temp#
	CALL Get.Token4
	SELECT CASE Number1#
	    CASE -1, 0, 1
		SELECT CASE Number2#
		    CASE -1, 0, 1
			Temp# = Number1# OR Number2#
			Last.Token = True
			EXIT SUB
		END SELECT
	END SELECT
	ERROR 5
    CASE "TXOR"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Number1# = Temp#
	CALL Get.Token
	CALL Parse1(Temp#)
	Number2# = Temp#
	CALL Get.Token4
	SELECT CASE Number1#
	    CASE -1, 0, 1
		SELECT CASE Number2#
		    CASE -1, 0, 1
			Temp# = Number1# XOR Number2#
			Last.Token = True
			EXIT SUB
		END SELECT
	END SELECT
	ERROR 5
	' date\time functions
    CASE "CLOCK"
	t1$ = FORMAT$(Now, "hh:mm:ss") ' seconds past midnight in dos
	t1! = CSNG(INT(VAL(MID$(t1$, 1, 2)))) * 3600! + CSNG(INT(VAL(MID$(t1$, 4, 2)))) * 60! + CSNG(INT(VAL(MID$(t1$, 7, 2))))
	Temp# = CDBL(t1!)
	Last.Token = True
    CASE "CLOCK$"
	Out3 = FORMAT$(Now, "ddd dd mmm yyyy\, hh:mma/p")
	Last.Token = False
    CASE "DATE$"
	Out3 = FORMAT$(Now, "mm-dd-yyyy")
	Last.Token = False
    CASE "TIME$"
	Out3 = FORMAT$(Now, "hh:mm:ss")
	Last.Token = False
	' conversion/calculation functions
    CASE "BIN$"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Out3 = Nul
	IF Temp# >= FalseD THEN
	    Digits = False
	    DO
		IF 2 ^ (Digits + 1) > Temp# THEN
		    EXIT DO
		END IF
		Digits = Digits + 1
	    LOOP
	    FOR Power = Digits TO 0 STEP -1
		IF Temp# - 2 ^ Power >= 0 THEN
		    Temp# = Temp# - 2 ^ Power
		    Out3 = Out3 + "1"
		ELSE
		    Out3 = Out3 + "0"
		END IF
	    NEXT
	END IF
	CALL Get.Token4
	Last.Token = False
    CASE "CHR$"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Out3 = CHR$(CINT(Temp#))
	CALL Get.Token4
	Last.Token = False
    CASE "HEX$"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Out3 = HEX$(Temp#)
	CALL Get.Token4
	Last.Token = False
    CASE "OCT$"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Out3 = OCT$(Temp#)
	CALL Get.Token4
	Last.Token = False
    CASE "STR$"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Out3 = STR$(Temp#)
	CALL Get.Token4
	Last.Token = False
    CASE "SUFFIX"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	CALL Suffix(Temp#, Out3)
	CALL Get.Token4
	Last.Token = False
    CASE "PARSESIZE$"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Out3 = ParseSize$(Temp#)
	CALL Get.Token4
	Last.Token = False
    CASE "OUTFILE$"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Out3 = UCASE$(Out3)
	CALL Get.Token4
	Last.Token = False
	Hidden = True
	CALL Out.File(Out3, -1)
    ' test function
    Case "CLOSE"
        Call Close.File
    ' memory functions
    CASE "POKE"
	CALL Poke.Value
    CASE "OUT"
	CALL Output.Port
    CASE "WAIT"
	CALL Wait.Port
    CASE "DEFSEG"
	CALL Defseg.Function
    CASE "ABSOLUTE"
	CALL Absolute.Function
    ' disk functions
    CASE "WARMBOOT"
	OUT &H64, &HF0
    CASE "COLDBOOT"
	DEF SEG = &HFFFF
	CALL Absolute(0)
	DEF SEG
    CASE "REBOOT"
	DEF SEG = &HFFFF
	CALL Absolute(0)
	DEF SEG
    CASE "CAPITALIZE"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	CALL Capitalize(Out3)
	CALL Get.Token4
	Last.Token = False
    CASE "KILL$", "DELETE$"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Out3 = UCASE$(Out3)
	CALL Get.Token4
	Last.Token = False
	KILL Out3
    CASE "CREATE$"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Out3 = UCASE$(Out3)
	CALL Get.Token4
	Last.Token = False
	CLOSE #1
	OPEN Out3 FOR OUTPUT AS #1
	PRINT #1, "";
	CLOSE #1
    CASE "FILEEXISTS"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	CALL Get.Bits(Var#)
	IF (Var# AND &H10) THEN
	   Temp# = 0
	ELSE
	   Var$ = DIR$(Out3)
	   IF Var$ = Nul THEN
	      Temp# = 0
	   ELSE
	      Temp# = -1
	   END IF
	END IF
	CALL Get.Token4
	Last.Token = True
    CASE "DIREXISTS"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	CALL Get.Bits(Var#)
	IF (Var# AND &H10) THEN
	   Temp# = -1
	ELSE
	   Temp# = 0
	END IF
	CALL Get.Token4
	Last.Token = True
    CASE "DRIVEEXISTS"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Var# = Temp#
	Temp# = 0#
	IF Last.Token THEN
	   Var = CINT(Var#)
	   IF Var >= 1 AND Var <= 26 THEN
	      CALL Check.Drive(Var, Temp#)
	   END IF
	ELSE
	   Var = ASC(UCASE$(Out3))
	   IF Var >= 65 AND Var <= 90 THEN
	      Var = Var - 64
	      CALL Check.Drive(Var, Temp#)
	   END IF
	END IF
	CALL Get.Token4
	Last.Token = True
    CASE "DIR$"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	IF LEN(Out3) THEN
	   Out3 = DIR$(Out3)
	ELSE
	   Out3 = DIR$
	END IF
	CALL Get.Token4
	Last.Token = False
    CASE "NAME", "RENAME"
	CALL Name.Expression
    CASE "CHDIR"
	CALL CD.Directory
    CASE "RMDIR"
	CALL RD.Directory
    CASE "MKDIR"
	CALL MD.Directory
    CASE "CURDIR$"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	IF Last.Token THEN
	   Var$ = CHR$(Temp# + 64) + ":"
	ELSE
	   Var$ = Out3
	END IF
	IF LEN(Var$) THEN
	   Out3 = CURDIR$(Var$)
	ELSE
	   Out3 = CURDIR$
	END IF
	CALL Get.Token4
	Last.Token = False
    CASE "DRIVE$"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	IF Last.Token THEN
	   Var$ = CHR$(Temp# + 64) + ":"
	ELSE
	   Var$ = Out3
	END IF
	IF LEN(Var$) THEN
	   Out3 = LEFT$(CURDIR$(Var$), 3)
	ELSE
	   Out3 = LEFT$(CURDIR$, 3)
	END IF
	CALL Get.Token4
	Last.Token = False
     CASE "FILEDATE"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	CALL Get.Token4
	Temp# = DFalse
	IF Last.Token = False THEN
	   IF LEN(Out3) THEN
	      CALL Filedate(Temp1$)
	      Out3 = Temp1$
	   END IF
	END IF
     CASE "FILETIME"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	CALL Get.Token4
	Temp# = DFalse
	IF Last.Token = False THEN
	   IF LEN(Out3) THEN
	      CALL Filetime(Temp1$)
	      Out3 = Temp1$
	   END IF
	END IF
     CASE "FILEATTR"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	CALL Get.Token4
	Temp# = DFalse
	IF Last.Token = False THEN
	   IF LEN(Out3) THEN
	      Last.Token = True
	      CALL Get.Bits(Temp#)
	   END IF
	END IF
     CASE "FILEBITS"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = DFalse
	IF Last.Token = False THEN
	   IF LEN(Out3) THEN
	      Last.Token = False
	      CALL Get.Bits(Temp#)
	      ' store filebits
	      Out3 = SPACE$(5)
	      IF (Temp# AND &H20) = &H20 THEN
		 MID$(Out3, 1, 1) = "A" ' archive
	      END IF
	      IF (Temp# AND &H10) = &H10 THEN
		 MID$(Out3, 2, 1) = "D" ' directory
	      END IF
	      IF (Temp# AND &H4) = &H4 THEN
		 MID$(Out3, 3, 1) = "S" ' system
	      END IF
	      IF (Temp# AND &H2) = &H2 THEN
		 MID$(Out3, 4, 1) = "H" ' hidden
	      END IF
	      IF (Temp# AND &H1) = &H1 THEN
		 MID$(Out3, 5, 1) = "R" ' read-only
	      END IF
	      Last.Token = False
	   END IF
	END IF
	CALL Get.Token4
    CASE "FILESIZE"
       CALL Get.Token3
       CALL Get.Token
       CALL Parse1(Temp#)
       CALL Get.Token4
       Temp# = DFalse
       IF Last.Token = False THEN
	  IF LEN(Out3) THEN
	     Last.Token = True
	     ASCIIZ = Out3 + CHR$(0)
	     ' restore directory search dta
	     InregsX.AX = &H1A00
	     InregsX.DS = VARSEG(DTAfile)
	     InregsX.DX = VARPTR(DTAfile)
	     CALL InterruptX(&H21, InregsX, OutregsX)
	     ' find first filename
	     InregsX.AX = &H4E00
	     InregsX.CX = &H27
	     InregsX.DS = VARSEG(ASCIIZ)
	     InregsX.DX = VARPTR(ASCIIZ)
	     CALL InterruptX(&H21, InregsX, OutregsX)
	     ' check findfirst error
	     IF (OutregsX.Flags AND &H1) = &H0 THEN
		' store file size
		Temp# = ASC(MID$(DTAfile.FileSize, 4, 1))
		Temp# = Temp# * &H100 + ASC(MID$(DTAfile.FileSize, 3, 1))
		Temp# = Temp# * &H100 + ASC(MID$(DTAfile.FileSize, 2, 1))
		Temp# = Temp# * &H100 + ASC(MID$(DTAfile.FileSize, 1, 1))
	      END IF
	      ' restore basic dta
	      InregsX.AX = &H1A00
	      InregsX.DS = BASIC.DTA.SEG
	      InregsX.DX = BASIC.DTA.OFF
	      CALL InterruptX(&H21, InregsX, OutregsX)
	   END IF
	END IF
     CASE "VSERIAL"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	CALL Get.Token4
	IF Last.Token THEN
	   Drive.Number% = CINT(Temp#)
	ELSE
	   Drive.Number% = CINT(ASC(UCASE$(Out3)) - 64)
	END IF
	Last.Token = False
	Out3 = Nul
	CALL Vserial(Out3, Drive.Number%)
     CASE "VTYPE"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	CALL Get.Token4
	IF Last.Token THEN
	   Drive.Number% = CINT(Temp#)
	ELSE
	   Drive.Number% = CINT(ASC(UCASE$(Out3)) - 64)
	END IF
	Last.Token = False
	Out3 = Nul
	CALL Vtype(Out3, Drive.Number%)
    CASE "VLABEL"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	CALL Get.Token4
	IF Last.Token THEN
	   Drive.Number% = CINT(Temp#)
	ELSE
	   Drive.Number% = CINT(ASC(UCASE$(Out3)) - 64)
	END IF
	Last.Token = False
	Out3 = Nul
	CALL Vlabel(Out3, Drive.Number%)
    CASE "SHELL$"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Var$ = Nul
	IF Last.Token = 0 THEN
	   Var$ = Out3
	END IF
	CALL Get.Token4
	CALL ClearStatusLine
	IF LEN(Var$) THEN
	   Var$ = Quote$ + Var$ + Quote$
	   SHELL Var$
	ELSE
	   SHELL
	END IF
	CALL DisplayStatusLine(Max.Row - 1, 1)
	Hidden = True
	Last.Token = False
    CASE "EDITFILE$"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Var$ = Nul
	IF Last.Token = 0 THEN
	   Var$ = Out3
	END IF
	CALL Get.Token4
	Var2$ = Editor$
	IF DIR$(Var2$) = "" THEN
	   COLOR Red, Black
	   PRINT Editor$; " not found."
	ELSE
	   CALL ClearStatusLine
	   IF LEN(Var$) THEN
	      Var2$ = Var2$ + " " + Quote$ + Var$ + Quote$
	      SHELL Var2$
	   ELSE
	      SHELL Var2$
	   END IF
	   CALL DisplayStatusLine(Max.Row - 1, 1)
	END IF
	Hidden = True
	Last.Token = False
    CASE "INTERPRETER$"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Var$ = Nul
	IF Last.Token = 0 THEN
	   Var$ = Out3
	END IF
	CALL Get.Token4
	Var2$ = Interpreter$
	IF DIR$(Var2$) = "" THEN
	   COLOR Red, Black
	   PRINT Interpreter$; " not found."
	ELSE
	   CALL ClearStatusLine
	   IF LEN(Var$) THEN
	      Var2$ = Var2$ + " " + Var$
	      SHELL Var2$
	   ELSE
	      SHELL Var2$
	   END IF
	   CALL DisplayStatusLine(Max.Row - 1, 1)
	END IF
	Hidden = True
	Last.Token = False
    CASE "CLS"
	CLS
	Out3 = ""
	Hidden = True
	Last.Token = False
	CALL StatusLine
    CASE "LCASE$"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Out3 = LCASE$(Out3)
	CALL Get.Token4
	Last.Token = False
    CASE "UCASE$"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Out3 = UCASE$(Out3)
	CALL Get.Token4
	Last.Token = False
    CASE "RTRIM$"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Out3 = RTRIM$(Out3)
	CALL Get.Token4
	Last.Token = False
    CASE "LTRIM$"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Out3 = LTRIM$(Out3)
	CALL Get.Token4
	Last.Token = False
    CASE "STRIM$"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Out3 = STRIM$(Out3)
	CALL Get.Token4
	Last.Token = False
    CASE "ENCRYPTX$"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	CALL Cryptograph(Out3)
	CALL Get.Token4
	Last.Token = False
    CASE "XTRIM$"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Out3 = XTRIM$(Out3)
	CALL Get.Token4
	Last.Token = False
    CASE "TRIM$"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Out3 = LTRIM$(RTRIM$(Out3))
	CALL Get.Token4
	Last.Token = False
    CASE "SPACE$"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Out3 = STRING$(Temp#, 32)
	CALL Get.Token4
	Last.Token = False
    CASE "MID$"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp4# = Temp#
	IF Strng = "," THEN
	    CALL Get.Token
	    CALL Parse1(Temp#)
	    Length# = Temp#
	    Out3 = MID$(Out3, Temp4#, Length#)
	ELSE
	    Out3 = MID$(Out3, Temp4#)
	END IF
	CALL Get.Token4
	Last.Token = False
    CASE "REMID$"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	CALL Get.Token
	CALL Parse1(Temp#)
	Start# = Temp#
	CALL Get.Token
	CALL Parse1(Temp#)
	Length# = Temp#
	Out3 = LEFT$(Out3, Start# - 1) + MID$(Out3, Start# + Length#)
	CALL Get.Token4
	Last.Token = False
    CASE "DEMID$"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	String1$ = Out3
	CALL Get.Token
	CALL Parse1(Temp#)
	IF LEN(String1$) > 0 AND LEN(Out3) > 0 THEN
	    Imbedded = INSTR(String1$, Out3)
	    WHILE Imbedded
		String1$ = LEFT$(String1$, Imbedded - 1) + MID$(String1$, Imbedded + LEN(Out3))
		Imbedded = INSTR(String1$, Out3)
	    WEND
	END IF
	Out3 = String1$
	CALL Get.Token4
	Last.Token = False
    CASE "INSTRX"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	String1$ = Out3
	CALL Get.Token
	CALL Parse1(Temp#)
	String2$ = Out3
	Temp# = 0#
	IF LEN(String1$) > 0 AND LEN(String2$) > 0 THEN
	   CALL InstrSUB(Var, String1$, String2$)
	   IF Var THEN
	      Temp# = CDBL(Var)
	   ELSE
	      Temp# = 0#
	   END IF
	END IF
	CALL Get.Token4
	Last.Token = True
    CASE "INSTRSUB"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	String1$ = Out3
	CALL Get.Token
	CALL Parse1(Temp#)
	String2$ = Out3
	Temp# = 0#
	IF LEN(String1$) > 0 AND LEN(String2$) > 0 THEN
	   CALL InstrSUB(Var, String1$, String2$)
	   IF Var THEN
	      Temp# = -1#
	   ELSE
	      Temp# = 0#
	   END IF
	END IF
	CALL Get.Token4
	Last.Token = True
    CASE "REPLACE$"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	String1$ = Out3
	CALL Get.Token
	CALL Parse1(Temp#)
	String2$ = Out3
	CALL Get.Token
	CALL Parse1(Temp#)
	String3$ = Out3
	IF String1$ = Nul OR String2$ = Nul OR String2$ = String3$ THEN
	    Out3 = String1$
	    EXIT SUB
	END IF
	Temp1 = 1
	DO
	    IF String1$ = Nul THEN
		EXIT DO
	    END IF
	    Var1 = INSTR(Temp1, String1$, String2$)
	    IF Var1 = False THEN
		EXIT DO
	    END IF
	    String1$ = LEFT$(String1$, Var1 - 1) + String3$ + MID$(String1$, Var1 + LEN(String2$))
	    Temp1 = Temp1 + LEN(String3$)
	LOOP
	Out3 = String1$
	CALL Get.Token4
	Last.Token = False
    CASE "LEFT$"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	CALL Get.Token
	CALL Parse1(Temp#)
	Out3 = LEFT$(Out3, Temp#)
	CALL Get.Token4
	Last.Token = False
    CASE "RIGHT$"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	CALL Get.Token
	CALL Parse1(Temp#)
	Out3 = RIGHT$(Out3, Temp#)
	CALL Get.Token4
	Last.Token = False
    CASE "STRING$"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Number# = Temp#
	CALL Get.Token
	CALL Parse1(Temp#)
	IF Last.Token THEN
	    Out3 = STRING$(Number#, Temp#)
	ELSE
	    Out3 = STRING$(Number#, Out3)
	END IF
	CALL Get.Token4
	Last.Token = False
    CASE "ENCRYPT$"
       Prime.Key = PrimeKey
       CALL Get.Token3
       CALL Get.Token
       CALL Parse1(Temp#)
       V$ = Out3
       Token3 = Last.Token
       IF Strng = "," THEN
	  CALL Get.Token
	  CALL Parse1(Temp2#)
	  Prime.Key = Temp2#
       END IF
       Out3 = V$
       IF Prime.Key >= 1 AND Prime.Key <= 255 THEN
	  IF Token3 THEN
	     Out3 = CHR$(Temp# XOR Prime.Key)
	  ELSE
	     FOR V = 1 TO LEN(Out3)
		MID$(Out3, V, 1) = CHR$(ASC(MID$(Out3, V, 1)) XOR Prime.Key)
	     NEXT
	  END IF
       ELSE
	  ERROR 5
       END IF
       CALL Get.Token4
       Last.Token = False
    CASE "FRE"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	IF Last.Token THEN
	    Temp# = FRE(Temp#)
	ELSE
	    Temp# = FRE(Out3)
	END IF
	CALL Get.Token4
	Last.Token = True
    CASE "TOGGLE"
	Hidden = True
	ToggleForm = NOT ToggleForm
    CASE "FORMAT$"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Number# = Temp#
	IF Strng = "," THEN
	    CALL Get.Token
	    CALL Parse1(Temp#)
	    IF Last.Token THEN
	       ERROR 5
	    ELSE
	       Out3 = FORMAT$(Number#, Out3)
	    END IF
	ELSE
	    Out3 = LTRIM$(STR$(Number#))
	END IF
	CALL Get.Token4
	Last.Token = False
    CASE "ABS"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = ABS(Temp#)
	CALL Get.Token4
	Last.Token = True
    CASE "ASC"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = ASC(Out3)
	CALL Get.Token4
	Last.Token = True
    CASE "EXP"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = EXP(Temp#)
	CALL Get.Token4
	Last.Token = True
    CASE "FBN" ' calculate xth fibonachi value
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Fibo.A# = 1
	Fibo.B# = 1
	IF Temp# >= 1 THEN
	    FOR Fibo.Number# = 1 TO Temp# - 1
		Fibo.Temp# = Fibo.B#
		Fibo.B# = Fibo.A# + Fibo.B#
		Fibo.A# = Fibo.Temp#
	    NEXT
	END IF
	Temp# = Fibo.B#
	CALL Get.Token4
	Last.Token = True
    CASE "FCT"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Factorial# = 1
	Factorial.Count# = INT(Temp#)
	IF Factorial.Count# >= 1 THEN
	    FOR Factorial.Count# = 1 TO Temp#
		Factorial# = Factorial# * Factorial.Count#
	    NEXT
	END IF
	Temp# = Factorial#
	CALL Get.Token4
	Last.Token = True
    CASE "FIX"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = FIX(Temp#)
	CALL Get.Token4
	Last.Token = True
    CASE "INP"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = INP(CINT(Temp#))
	CALL Get.Token4
	Last.Token = True
    CASE "INSTR"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	IF Last.Token THEN
	    Start# = Temp#
	    CALL Get.Token
	    CALL Parse1(Temp#)
	    Stored.String$ = Out3
	    CALL Get.Token
	    CALL Parse1(Temp#)
	ELSE
	    Start# = 1#
	    Stored.String$ = Out3
	    CALL Get.Token
	    CALL Parse1(Temp#)
	END IF
	Temp# = INSTR(Start#, Stored.String$, Out3)
	CALL Get.Token4
	Last.Token = True
    CASE "INT"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = INT(Temp#)
	CALL Get.Token4
	Last.Token = True
    CASE "LEN"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = LEN(Out3)
	CALL Get.Token4
	Last.Token = True
    CASE "LOG"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = LOG(Temp#)
	CALL Get.Token4
	Last.Token = True
    CASE "ISEVEN"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = (Temp# MOD 2#) = 0
	CALL Get.Token4
	Last.Token = True
    CASE "ISODD"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = (Temp# MOD 2#) = 1
	CALL Get.Token4
	Last.Token = True
    CASE "ISPRM" ' calculate x is prime
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Prime# = INT(Temp#)
	Prime.Flag = True
	IF Prime# > 2# THEN
	    FOR Loop.Count# = 2# TO INT(SQR(Prime#))
		IF StdinEnabled = 0 THEN
		   IF INKEY$ = CHR$(3) THEN EXIT FOR
		END IF
		IF Prime# / Loop.Count# = INT(Prime# / Loop.Count#) THEN
		    Prime.Flag = False
		    EXIT FOR
		END IF
	    NEXT
	END IF
	IF Prime.Flag = False THEN
	    Temp# = FalseD
	ELSE
	    Temp# = TrueD
	END IF
	CALL Get.Token4
	Last.Token = True
    CASE "PRMIS" ' calculate which x is prime
       CALL Get.Token3
       CALL Get.Token
       CALL Parse1(Temp#)
       Prime# = INT(Temp#)
       Prime.Count# = 0#
       IF Prime# > 1# THEN
	  FOR Prime2# = 2# TO Prime#
	     IF StdinEnabled = 0 THEN
		IF INKEY$ = CHR$(3) THEN EXIT FOR
	     END IF
	     Prime.Flag = True
	     FOR Loop.Count# = 2# TO INT(SQR(Prime2#))
		IF Prime2# / Loop.Count# = INT(Prime2# / Loop.Count#) THEN
		   Prime.Flag = False
		   EXIT FOR
		END IF
	     NEXT
	     IF Prime.Flag THEN
		Prime.Count# = Prime.Count# + 1#
	     END IF
	  NEXT
       END IF
       Temp# = Prime.Count#
       CALL Get.Token4
       Last.Token = True
    CASE "PRF" ' calculate xth perfect number
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	CALL Get.Token4
	Last.Token = True
	Perfect# = INT(Temp#)
	SELECT CASE Perfect#
	    CASE 1#
		x# = 1#
	    CASE 2#
		x# = 2#
	    CASE 3#
		x# = 4#
	    CASE 4#
		x# = 6#
	    CASE 5#
		x# = 12#
	    CASE 6#
		x# = 16#
	    CASE 7#
		x# = 18#
		'CASE 8# ' overflows beyond double precision
		'  X# = 30#
	    CASE ELSE
		ERROR 130
	END SELECT
	' equation for perfect number
	Temp# = 2 ^ x# * (2 ^ (x# + 1) - 1)
	Last.Token = True
    CASE "PRM" ' calculate xth prime
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Prime.Number# = INT(Temp#)
	Prime# = 1
	IF Prime.Number# > FalseD THEN
	    Prime.Counter# = False
	    StartTime2# = Timeit!
	    DO
		IF StdinEnabled = 0 THEN
		   IF INKEY$ = CHR$(3) THEN EXIT DO
		END IF
		Prime# = Prime# + 1
		Prime.Flag = False
		FOR Factor# = 2# TO INT(SQR(Prime#))
		    IF StdinEnabled = 0 THEN
		       IF INKEY$ = CHR$(3) THEN EXIT DO
		    END IF
		    RunTime2# = Timeit! - StartTime2#
		    IF RunTime2# < 0# THEN RunTime2# = RunTime2# + 86400#
		    IF RunTime2# >= 1# THEN
		       StartTime2# = Timeit!
		       Var1 = CSRLIN
		       Var2 = POS(0)
		       CALL DisplayStatusLine(Var1, Var2)
		    END IF
		    IF Prime# / Factor# = INT(Prime# / Factor#) THEN
			Prime.Flag = True
			EXIT FOR
		    END IF
		NEXT
		IF Prime.Flag = False THEN
		    Prime.Counter# = Prime.Counter# + 1
		END IF
		IF Prime.Counter# = Prime.Number# THEN
		    EXIT DO
		END IF
	    LOOP
	END IF
	Temp# = Prime#
	CALL Get.Token4
	Last.Token = True
    CASE "SGN"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = SGN(Temp#)
	CALL Get.Token4
	Last.Token = True
    CASE "SQR"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = SQR(Temp#)
	CALL Get.Token4
	Last.Token = True
    CASE "CBR"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = Temp# ^ (1# / 3#)
	CALL Get.Token4
	Last.Token = True
    CASE "GCM"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Number1# = Temp#
	CALL Get.Token
	CALL Parse1(Temp#)
	Number2# = Temp#
	' find last factor which divides into both numbers
	Factor1# = 1
	Number3# = Number1#
	FOR Factor2# = 2 TO Number1#
	    DO
		IF Number3# / Factor2# = Number3# \ Factor2# THEN
		    IF Number2# / Factor2# = Number2# \ Factor2# THEN
			IF Factor2# > Factor1# THEN
			    Factor1# = Factor2#
			END IF
		    END IF
		    Number3# = Number3# / Factor2#
		ELSE
		    EXIT DO
		END IF
	    LOOP
	NEXT
	Temp# = Factor1#
	CALL Get.Token4
	Last.Token = True
    CASE "LCD"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Number1# = Temp#
	CALL Get.Token
	CALL Parse1(Temp#)
	Number2# = Temp#
	' find first factor which divides into both numbers
	Factor1# = 0
	FOR Factor2# = 2 TO Number1#
	    IF Number1# / Factor2# = Number1# \ Factor2# THEN
		IF Number2# / Factor2# = Number2# \ Factor2# THEN
		    Factor1# = Factor2#
		    EXIT FOR
		END IF
	    END IF
	NEXT
	IF Factor1# = 0# THEN
	    Factor1# = 1#
	END IF
	IF Factor1# = Number2# THEN
	    Factor1# = 1#
	END IF
	Temp# = Factor1#
	CALL Get.Token4
	Last.Token = True
    CASE "NTH"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Number# = Temp#
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = Number# ^ (1# / Temp#)
	CALL Get.Token4
	Last.Token = True
    CASE "PWR"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Number# = Temp#
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = Number# ^ Temp#
	CALL Get.Token4
	Last.Token = True
    CASE "VAL"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = VAL(Out3)
	CALL Get.Token4
	Last.Token = True
    CASE "PEEK"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = PEEK(Temp#)
	CALL Get.Token4
	Last.Token = True

	REM Basic trigonometric functions..

    CASE "ATN"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = ATN(Temp#)
	CALL Get.Token4
	Last.Token = True
    CASE "COS"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = COS(Temp#)
	CALL Get.Token4
	Last.Token = True
    CASE "COT"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = COS(Temp#) / SIN(Temp#)
	CALL Get.Token4
	Last.Token = True
    CASE "CSC"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = 1# / SIN(Temp#)
	CALL Get.Token4
	Last.Token = True
    CASE "SEC"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = 1# / COS(Temp#)
	CALL Get.Token4
	Last.Token = True
    CASE "SIN"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = SIN(Temp#)
	CALL Get.Token4
	Last.Token = True
    CASE "TAN"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = TAN(Temp#)
	CALL Get.Token4
	Last.Token = True

	REM Remaining nonintrinsic trigonometric functions..
	'Secant
    CASE "SECANT"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = 1# / COS(Temp#)
	CALL Get.Token4
	Last.Token = True
	'Cosecant
    CASE "COSEC"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = 1# / SIN(Temp#)
	CALL Get.Token4
	Last.Token = True
	'Cotangent
    CASE "COTAN"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = 1# / TAN(Temp#)
	CALL Get.Token4
	Last.Token = True
	'Inverse Cosine
    CASE "ARCCOS"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = ATN(-Temp# / SQR(-Temp# * Temp# + 1)) + 2 * ATN(1)
	CALL Get.Token4
	Last.Token = True
	'Inverse Cosecant
    CASE "ARCCOSEC"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = ATN(Temp# / SQR(Temp# * Temp# - 1)) + (SGN(Temp#) - 1) * (2 * ATN(1))
	CALL Get.Token4
	Last.Token = True
	'Inverse Cotangent
    CASE "ARCCOTAN"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = ATN(Temp#) + 2 * ATN(1)
	CALL Get.Token4
	Last.Token = True
	'Inverse Secant
    CASE "ARCSEC"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = ATN(Temp# / SQR(Temp# * Temp# - 1)) + SGN((Temp#) - 1) * (2 * ATN(1))
	CALL Get.Token4
	Last.Token = True
	'Inverse Sine
    CASE "ARCSIN"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = ATN(Temp# / SQR(-Temp# * Temp# + 1))
	CALL Get.Token4
	Last.Token = True
	'Inverse Hyperbolic Cosine
    CASE "HARCCOS"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = LOG(Temp# + SQR(Temp# * Temp# - 1))
	CALL Get.Token4
	Last.Token = True
	'Inverse Hyperbolic Cosecant
    CASE "HARCCOSEC"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = LOG((SGN(Temp#) * SQR(Temp# * Temp# + 1) + 1) / Temp#)
	CALL Get.Token4
	Last.Token = True
	'Inverse Hyperbolic Cotangent
    CASE "HARCCOTAN"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = LOG((Temp# + 1) / (Temp# - 1)) / 2
	CALL Get.Token4
	Last.Token = True
	'Inverse Hyperbolic Secant
    CASE "HARCSEC"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = LOG((SQR(-Temp# * Temp# + 1) + 1) / Temp#)
	CALL Get.Token4
	Last.Token = True
	'Inverse Hyperbolic Sine
    CASE "HARCSIN"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = LOG(Temp# + SQR(Temp# * Temp# + 1))
	CALL Get.Token4
	Last.Token = True
	'Inverse Hyperbolic Tangent
    CASE "HARCTAN"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = LOG((1 + Temp#) / (1 - Temp#)) / 2
	CALL Get.Token4
	Last.Token = True
	'Hyperbolic Cosine
    CASE "HCOS"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = (EXP(Temp#) + EXP(-Temp#)) / 2
	CALL Get.Token4
	Last.Token = True
	'Hyperbolic Cosecant
    CASE "HCOSEC"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = 2 / (EXP(Temp#) - EXP(-Temp#))
	CALL Get.Token4
	Last.Token = True
	'Hyperbolic Cotangent
    CASE "HCOTAN"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = (EXP(Temp#) + EXP(-Temp#)) / (EXP(Temp#) - EXP(-Temp#))
	CALL Get.Token4
	Last.Token = True
	'Hyperbolic Secant
    CASE "HSEC"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = 2 / (EXP(Temp#) + EXP(-Temp#))
	CALL Get.Token4
	Last.Token = True
	'Hyperbolic Sine
    CASE "HSIN"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = (EXP(Temp#) - EXP(-Temp#)) / 2
	CALL Get.Token4
	Last.Token = True
	'Hyperbolic Tangent
    CASE "HTAN"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = (EXP(Temp#) - EXP(-Temp#)) / (EXP(Temp#) + EXP(-Temp#))
	CALL Get.Token4
	Last.Token = True

	'Convert Degrees to Radians
    CASE "DEGTORAD"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = (PI# / 180) * Temp#
	CALL Get.Token4
	Last.Token = True
	'Convert Radians to Degrees
    CASE "RADTODEG"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = (180 / PI#) * Temp#
	CALL Get.Token4
	Last.Token = True

	' Hypotenuse
    CASE "HYPOT"
	CALL Get.Token3
	CALL Get.Token
	CALL Parse1(Temp#)
	Number# = Temp#
	CALL Get.Token
	CALL Parse1(Temp#)
	Temp# = SQR(Number# * Number# + Temp# * Temp#)
	CALL Get.Token4
	Last.Token = True
    CASE ELSE
	ERROR 92
END SELECT
END SUB

SUB Parse.Command.Line (V$)
 CALL Strip.Semicolon(V$)
 ' parse command line
 DO
    V = INSTR(V$, ";")
    IF V THEN
	Out2 = LEFT$(V$, V - 1)
	V$ = MID$(V$, V + 1)
	Out2 = LTRIM$(RTRIM$(Out2))
	IF LEN(Out2) THEN
	   COLOR White, Black
	   PRINT Out2; " ";
	   CALL Remove.Spaces
	   Recurse = 0
	   Hidden = False
	   CALL Enter.Equate
	   IF Hidden THEN
	      PRINT "Equals: ";
	      COLOR Green
	      PRINT "<nul assign>"
	   ELSE
	      IF Assign THEN
		 PRINT "Equals: ";
		 COLOR Red
		 PRINT "<assign>"
	      END IF
	   END IF
	END IF
     ELSE
	Out2 = V$
	Out2 = LTRIM$(RTRIM$(Out2))
	IF LEN(Out2) THEN
	   COLOR White, Black
	   PRINT Out2; " ";
	   CALL Remove.Spaces
	   Recurse = 0
	   Hidden = False
	   CALL Enter.Equate
	   IF Hidden THEN
	      PRINT "Equals: ";
	      COLOR Green
	      PRINT "<nul assign>"
	   ELSE
	      IF Assign THEN
		 PRINT "Equals: ";
		 COLOR Red
		 PRINT "<assign>"
	      END IF
	   END IF
	END IF
	EXIT DO
     END IF
 LOOP
END SUB

' process a number
SUB Parse.Numeric (Temp#)
SELECT CASE RIGHT$(UCASE$(Strng), 1) ' compare character after number
    CASE "Q" ' perfect number suffix
	FOR Digit = 1 TO LEN(Strng) - 1
	    OutX$ = UCASE$(MID$(Strng, Digit, 1))
	    SELECT CASE OutX$
		CASE "0", "1", "2", "3", "4", "5", "6", "7", "8", "9"
		    ' nul
		CASE ELSE
		    ERROR 130
	    END SELECT
	NEXT
	Temp# = CDBL(VAL(LEFT$(Strng, LEN(Strng) - 1)))
	Perfect# = 0#
	Perfect.Number# = INT(Temp#)
	IF ABS(Perfect.Number#) > 7# THEN
	    ERROR 130
	END IF
	IF Perfect.Number# > FalseD THEN
	    SELECT CASE Perfect.Number#
		CASE 1#
		    x# = 1#
		CASE 2#
		    x# = 2#
		CASE 3#
		    x# = 4#
		CASE 4#
		    x# = 6#
		CASE 5#
		    x# = 12#
		CASE 6#
		    x# = 16#
		CASE 7#
		    x# = 18#
		    'CASE 8# ' overflows beyond double precision
		    '  X# = 30#
		CASE ELSE
		    ERROR 130
	    END SELECT
	    ' equation for perfect number
	    Perfect# = 2 ^ x# * (2 ^ (x# + 1) - 1)
	END IF
	Temp# = Perfect#
    CASE "P" ' prime number suffix
	FOR Digit = 1 TO LEN(Strng) - 1
	    OutX$ = UCASE$(MID$(Strng, Digit, 1))
	    SELECT CASE OutX$
		CASE "0", "1", "2", "3", "4", "5", "6", "7", "8", "9"
		    ' nul
		CASE ELSE
		    ERROR 130
	    END SELECT
	NEXT
	Temp# = CDBL(VAL(LEFT$(Strng, LEN(Strng) - 1)))
	Prime.Number# = INT(Temp#)
	Prime# = 1
	IF Prime.Number# > False THEN
	    Prime.Counter# = False
	    StartTime2# = Timeit!
	    DO
		IF StdinEnabled = 0 THEN
		   IF INKEY$ = CHR$(3) THEN EXIT DO
		END IF
		RunTime2# = Timeit! - StartTime2#
		IF RunTime2# < 0# THEN RunTime2# = RunTime2# + 86400#
		IF RunTime2# >= 1# THEN
		   StartTime2# = Timeit!
		   Var1 = CSRLIN
		   Var2 = POS(0)
		   CALL DisplayStatusLine(Var1, Var2)
		END IF
		Prime# = Prime# + 1
		Prime.Flag = False
		FOR Factor# = 2# TO INT(SQR(Prime#))
		    IF StdinEnabled = 0 THEN
		       IF INKEY$ = CHR$(3) THEN EXIT DO
		    END IF
		    IF Prime# / Factor# = INT(Prime# / Factor#) THEN
			Prime.Flag = True
			EXIT FOR
		    END IF
		NEXT
		IF Prime.Flag = False THEN
		    Prime.Counter# = Prime.Counter# + 1
		END IF
		IF Prime.Counter# = Prime.Number# THEN
		    EXIT DO
		END IF
	    LOOP
	END IF
	Temp# = Prime#
    CASE "F" ' factorial
	Temp2 = False
	Temp3 = LEN(Strng)
	DO
	    IF UCASE$(MID$(Strng, Temp3, 1)) <> "F" THEN
		EXIT DO
	    END IF
	    Temp2 = Temp2 + 1
	    Temp3 = Temp3 - 1
	LOOP
	FOR Decimal.Digit = 1 TO Temp3
	    OutX$ = UCASE$(MID$(Strng, Decimal.Digit, 1))
	    SELECT CASE OutX$
		CASE "0", "1", "2", "3", "4", "5", "6", "7", "8", "9"
		    ' nul
		CASE ELSE
		    ERROR 130
	    END SELECT
	NEXT
	Temp4# = VAL(LEFT$(Strng, Temp3))
	FOR Number = 1 TO Temp2
	    Temp# = 1#
	    FOR Factorial# = 2# TO Temp4#
		Temp# = Temp# * Factorial#
	    NEXT
	    Temp4# = Temp#
	NEXT
    CASE "H" ' hexidecimal suffix
	Temp# = False
	Hex.Power = False
	FOR Hex.Digit = LEN(Strng) - 1 TO 1 STEP -1
	    OutX$ = UCASE$(MID$(Strng, Hex.Digit, 1))
	    SELECT CASE OutX$
		CASE "0" TO "9", "A" TO "F"
		    Hex.Value = VAL("&H" + OutX$)
		    Temp# = Temp# + Hex.Value * 16 ^ Hex.Power
		    Hex.Power = Hex.Power + 1
		CASE ELSE
		    ERROR 130
	    END SELECT
	NEXT
    CASE "O" ' octal suffix
	Temp# = False
	Octal.Power = False
	FOR Octal.Digit = LEN(Strng) - 1 TO 1 STEP -1
	    OutX$ = MID$(Strng, Octal.Digit, 1)
	    IF OutX$ >= "0" AND OutX$ <= "7" THEN
		Octal.Value = VAL(OutX$)
		Temp# = Temp# + Octal.Value * 8 ^ Octal.Power
		Octal.Power = Octal.Power + 1
	    ELSE
		ERROR 130
	    END IF
	NEXT
    CASE "B" ' binary suffix
	Temp# = False
	Binary.Power = False
	FOR Binary.Digit = LEN(Strng) - 1 TO 1 STEP -1
	    OutX$ = MID$(Strng, Binary.Digit, 1)
	    SELECT CASE OutX$
		CASE "0"
		    ' nul
		CASE "1"
		    Temp# = Temp# + 2 ^ Binary.Power
		CASE ELSE
		    ERROR 130
	    END SELECT
	    Binary.Power = Binary.Power + 1
	NEXT
    CASE "0" TO "9", "D", "E" ' decimal/exponent suffix
	Decimal = False
	Exponent = False
	Number$ = Strng
Start.Token:
	FOR Decimal.Digit = 1 TO LEN(Number$)
	    OutX$ = UCASE$(MID$(Number$, Decimal.Digit, 1))
	    SELECT CASE OutX$
		CASE "0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "-", "+"
		    ' nul
		CASE "D", "E"
		    IF Exponent THEN
			ERROR 140
		    END IF
		    Exponent = True
		    Exponent$ = UCASE$(RIGHT$(Number$, 1))
		    IF Exponent$ = "D" OR Exponent$ = "E" THEN
			' concatenate pre-parsed tokens into exponent
			Unary$ = MID$(Out2, Token.Index, 1)
			IF Unary$ = "+" OR Unary$ = "-" THEN
			    Var$ = LEFT$(Out2, Token.Index - 1)
			    CALL Get.Token
			    Out2 = Var$ + Strng + MID$(Out2, Token.Index)
			    CALL Get.Token
			    IF Strng = Nul THEN
				ERROR 140
			    END IF
			    Number$ = Number$ + Unary$ + Strng
			    IF Strng = "-" OR Strng = "+" THEN
				ERROR 140
			    END IF
			    IF INSTR(Strng, ".") THEN
				ERROR 140
			    END IF
			    DO
				IF LEFT$(Number$, 1) = "-" THEN
				    Number$ = MID$(Number$, 2)
				ELSE
				    IF LEFT$(Number$, 1) = "+" THEN
					Number$ = MID$(Number$, 2)
				    ELSE
					EXIT DO
				    END IF
				END IF
			    LOOP
			    Decimal = False
			    Exponent = False
			    GOTO Start.Token
			ELSE
			    ERROR 140
			END IF
		    END IF
		CASE "."
		    IF Exponent THEN
			ERROR 140
		    END IF
		    IF Decimal THEN
			ERROR 140
		    END IF
		    Decimal = True
		CASE ELSE
		    ERROR 130
	    END SELECT
	NEXT
	Temp# = CDBL(VAL(Number$))
    CASE ELSE
	ERROR 130
END SELECT
CALL Get.Token
Last.Token = True
END SUB

' parse string in quotes
SUB Parse.Quoted
Out3 = Nul
DO UNTIL MID$(Out2, Token.Index, 1) = Quote$
    Out3 = Out3 + MID$(Out2, Token.Index, 1)
    Token.Index = Token.Index + 1
    IF Token.Index > LEN(Out2) THEN
	Strng = "<extra/missing quotes>"
	ERROR 92
	EXIT DO
    END IF
LOOP
CALL Get.Token
CALL Get.Token
Last.Token = False
END SUB

' logical parser
SUB Parse1 (Temp#)
CALL Parse2(Temp#) ' get next operator precedence
Token.Parsed$ = Strng ' store token
' process token
DO
    SELECT CASE Token.Parsed$
	CASE "|", "&", "~", "?", ":", "#", "@", "`"
	    ' Nul
	CASE ELSE
	    EXIT DO
    END SELECT
    Token.Stored$ = Out3 ' store current token
    CALL Get.Token ' read next token
    CALL Parse2(Temp2#) ' get next operator
    Out4 = Out3 ' reset current string
    Out3 = Token.Stored$ ' restore previous string
    CALL Arith(Token.Parsed$, Temp#, Temp2#) ' calculate expression
    Token.Parsed$ = Strng ' store next token
LOOP
END SUB

' relational parser
SUB Parse2 (Temp#)
CALL Parse3(Temp#) ' get next operator precedence
Token.Parsed$ = Strng ' store token
' process token
DO
    SELECT CASE Token.Parsed$
	CASE "<", ">", "=", ">=", "<=", "<>", "=="
	    ' Nul
	CASE ELSE
	    EXIT DO
    END SELECT
    Token.Stored$ = Out3 ' store current token
    CALL Get.Token ' read next token
    CALL Parse3(Temp2#) ' get next operator
    Out4 = Out3 ' reset current string
    Out3 = Token.Stored$ ' restore previous string
    CALL Arith(Token.Parsed$, Temp#, Temp2#) ' calculate expression
    Token.Parsed$ = Strng ' store next token
LOOP
END SUB

' addition/subtraction parser
SUB Parse3 (Temp#)
CALL Parse4(Temp#) ' get next operator precedence
Token.Parsed$ = Strng ' store token
' process token
DO
    SELECT CASE Token.Parsed$
	CASE "+", "-"
	    ' Nul
	CASE ELSE
	    EXIT DO
    END SELECT
    Token.Stored$ = Out3 ' store current token
    CALL Get.Token ' read next token
    CALL Parse4(Temp2#) ' get next operator
    Out4 = Out3 ' reset current string
    Out3 = Token.Stored$ ' restore previous string
    CALL Arith(Token.Parsed$, Temp#, Temp2#) ' calculate expression
    Token.Parsed$ = Strng ' store next token
LOOP
END SUB

' modulo parser
SUB Parse4 (Temp#)
CALL Parse5(Temp#) ' get next operator precedence
Token.Parsed$ = Strng ' store token
' process token
DO
    SELECT CASE Token.Parsed$
	CASE "%"
	    ' Nul
	CASE ELSE
	    EXIT DO
    END SELECT
    Token.Stored$ = Out3 ' store current token
    CALL Get.Token ' read next token
    CALL Parse5(Temp2#) ' get next operator
    Out4 = Out3 ' reset current string
    Out3 = Token.Stored$ ' restore previous string
    CALL Arith(Token.Parsed$, Temp#, Temp2#) ' calculate expression
    Token.Parsed$ = Strng ' store next token
LOOP
END SUB

' integer division parser
SUB Parse5 (Temp#)
CALL Parse6(Temp#) ' get next operator precedence
Token.Parsed$ = Strng ' store token
' process token
DO
    SELECT CASE Token.Parsed$
	CASE "\"
	    ' Nul
	CASE ELSE
	    EXIT DO
    END SELECT
    Token.Stored$ = Out3 ' store current token
    CALL Get.Token ' read next token
    CALL Parse6(Temp2#) ' get next operator
    Out4 = Out3 ' reset current string
    Out3 = Token.Stored$ ' restore previous string
    CALL Arith(Token.Parsed$, Temp#, Temp2#) ' calculate expression
    Token.Parsed$ = Strng ' store next token
LOOP
END SUB

' multiplication/division parser
SUB Parse6 (Temp#)
CALL Parse7(Temp#) ' get next operator precedence
Token.Parsed$ = Strng ' store token
' process token
DO
    SELECT CASE Token.Parsed$
	CASE "*", "/"
	    ' Nul
	CASE ELSE
	    EXIT DO
    END SELECT
    Token.Stored$ = Out3 ' store current token
    CALL Get.Token ' read next token
    CALL Parse7(Temp2#) ' get next operator
    Out4 = Out3 ' reset current string
    Out3 = Token.Stored$ ' restore previous string
    CALL Arith(Token.Parsed$, Temp#, Temp2#) ' calculate expression
    Token.Parsed$ = Strng ' store next token
LOOP
END SUB

' power parser
SUB Parse7 (Temp#)
CALL Parse7a(Temp#) ' get next operator precedence
Token.Parsed$ = Strng ' store token
' process token
DO
    SELECT CASE Token.Parsed$
	CASE "^", "^=", "^-=", "^+="
	    ' Nul
	CASE ELSE
	    EXIT DO
    END SELECT
    Token.Stored$ = Out3 ' store current token
    CALL Get.Token ' read next token
    CALL Parse7a(Temp2#) ' get next operator
    Out4 = Out3 ' reset current string
    Out3 = Token.Stored$ ' restore previous string
    CALL Arith(Token.Parsed$, Temp#, Temp2#) ' calculate expression
    Token.Parsed$ = Strng ' store next token
LOOP
END SUB

' dual-unary parser
SUB Parse7a (Temp#)
CALL Parse8(Temp#) ' get next operator precedence
Token.Parsed$ = Strng ' store token
' process token
DO
    SELECT CASE Token.Parsed$
	CASE "<<", ">>", "--", "++", "**", "//", "||", "##", "^^", "=="
	    ' Nul
	CASE ELSE
	    EXIT DO
    END SELECT
    Token.Stored$ = Out3 ' store current token
    CALL Get.Token ' read next token
    Out4 = Out3 ' reset current string
    Out3 = Token.Stored$ ' restore previous string
    CALL Arith(Token.Parsed$, Temp#, Temp2#) ' calculate expression
    Token.Parsed$ = Strng ' store next token
LOOP
END SUB

' not/unary plus/unary negative parser
SUB Parse8 (Temp#)
Token.Negate$ = Nul ' reset not/unary tokens
Token.Parsed$ = Strng ' store token
' process token
DO
    SELECT CASE Token.Parsed$
	CASE "!", "-", "+", "--", "++"
	    ' Nul
	CASE ELSE
	    EXIT DO
    END SELECT
    CALL Get.Token ' read next token
    ' store combined not/unary tokens
    Token.Negate$ = Token.Negate$ + Token.Parsed$
    Token.Parsed$ = Strng
LOOP
CALL Parse9(Temp#) ' get next operator precedence
' process the combined operators in reverse
FOR Token.Type = LEN(Token.Negate$) TO 1 STEP -1
    SELECT CASE MID$(Token.Negate$, Token.Type, 1)
	CASE "+"
	    ' nul calculation for unary plus
	CASE "-"
	    Temp# = -Temp# ' perform negate
	CASE "!" ' not
	    Temp# = NOT Temp# ' perform not calculation
    END SELECT
NEXT
END SUB

' string/numeric expression parser,
' routine returns calculated expression string/value,
' calls Parse1 recursively for values inside parenthesis/functions.
SUB Parse9 (Temp#)
SELECT CASE Token ' determine token type
    CASE 0 ' unknown token
	IF Last.Token = 3 THEN
	    Strng = "<extra token>"
	    ERROR 92
	END IF
	IF Strng = Nul THEN
	    Strng = "<missing token>"
	END IF
	ERROR 92
	Last.Token = False
    CASE 1 ' token is symbol
	Token.Parsed$ = Strng
	SELECT CASE Token.Parsed$ ' determine token
	    CASE ",", ";" ' separaters
		Token.Index = Token.Index + 1
		Last.Token = 3
	    CASE Quote$ ' parse string
		CALL Parse.Quoted
		Quotes = True
	    CASE "(" ' calculate opening parenthesis
		CALL Get.Token ' read next token value inside parenthesis
		IF Strng = ")" THEN
		    Strng = "<empty closing token>"
		    ERROR 92
		END IF
		IF Strng <> ")" THEN
		    DO ' calculate value
			CALL Parse1(Temp#) ' call parse entry
		    LOOP UNTIL Strng = ")" OR Token = False ' check closing parenthesis
		END IF
		IF Token = False THEN
		    Strng = "<missing closing token>"
		    ERROR 92
		END IF
		CALL Get.Token ' read next token after parenthesis
	    CASE "["
		CALL Get.Token
		IF Strng = "]" THEN
		    Strng = "<empty closing token>"
		    ERROR 92
		END IF
		IF Strng <> "]" THEN
		    DO
			CALL Parse1(Temp#)
		    LOOP UNTIL Strng = "]" OR Token = False
		END IF
		IF Token = False THEN
		    Strng = "<missing closing token>"
		    ERROR 92
		END IF
		CALL Get.Token
	    CASE "{"
		CALL Get.Token
		IF Strng = "}" THEN
		    Strng = "<empty closing token>"
		    ERROR 92
		END IF
		IF Strng <> "}" THEN
		    DO
			CALL Parse1(Temp#)
		    LOOP UNTIL Strng = "}" OR Token = False
		END IF
		IF Token = False THEN
		    Strng = "<missing closing token>"
		    ERROR 92
		END IF
		CALL Get.Token
	    CASE ")", "]", "}" ' check token is closing parenthesis
		Strng = "<extra closing token>"
		ERROR 92
	END SELECT
    CASE 2 ' token is numeric value
	CALL Parse.Numeric(Temp#)
    CASE 3 ' token type is alphabetic
	SELECT CASE LEN(Strng) ' check variable type
	    CASE 1
		CALL Parse.Alphabetic1(Temp#)
	    CASE 2
		CALL Parse.Alphabetic2(Temp#)
	    CASE ELSE
		CALL Parse.Alphabetic3(Temp#)
	END SELECT
	CALL Get.Token
END SELECT
END SUB

' parses FN(<x>)
SUB ParseFN (Temp#)
Var$ = LEFT$(Out2, Token.Index - 3)
CALL Get.Token3
CALL Get.Token
CALL Parse1(Temp4#)
CALL Get.Token4
Func% = CINT(Temp4#)
IF Func% >= 1 AND Func% <= Max.FNs THEN
    Var2$ = MID$(Out2, Token.Index)
    Out2 = Definitions(Func%)
    Out2 = LTRIM$(Out2)
    ' test for nul token
    IF Out2 = Nul THEN
	Strng = "<bad FN token>"
	ERROR 92
    END IF
    ' count recursing calls
    Recurse = Recurse + 1
    IF Recurse >= Max.Recurse THEN
	Strng = "<bad FN recurse:" + STR$(Recurse) + " levels>"
	ERROR 92
    END IF
    ' recursively call functions
    Token.Index = 1
    CALL Get.Token
    CALL Parse1(Temp#)
    ' test huge recursed FN
    IF LEN(Var$) + LEN(Var2$) >= Max.FN.Length THEN
	Var$ = Nul
	Var2$ = Nul
	Strng = "<bad FN string>"
	ERROR 92
    END IF
    Out2 = Var$ + Var2$
    Token.Index = LEN(Var$) + 1
    EXIT SUB
END IF
Strng = "<bad FN range>"
ERROR 92
END SUB

' parses FZ(<x>)
SUB ParseFZ (Temp#)
CALL Get.Token3
CALL Get.Token
CALL Parse1(Temp4#)
CALL Get.Token4
Func% = CINT(Temp4#)
IF Func% >= 1 AND Func% <= Max.FNs THEN
    Out3 = Definitions(Func%)
    Out3 = LTRIM$(Out3)
    Out3 = RTRIM$(Out3)
    IF Out3 = Nul THEN
	Strng = "<bad FZ token>"
	ERROR 92
    END IF
    Last.Token = False
    EXIT SUB
END IF
Strng = "<bad FZ range>"
ERROR 92
END SUB

' formats bytes to a string
FUNCTION ParseSize$ (Temp#)
'Metric Suffixes:
'================
'Byte  (B)  - 1024^0  '  00x - FFx
'Kilo  (K)  - 1024^1
'Mega  (M)  - 1024^2
'Giga  (G)  - 1024^3
'Tera  (T)  - 1024^4
'Peta  (P)  - 1024^5
'Exa   (E)  - 1024^6
'Zetta (Z)  - 1024^7
'Yotta (Y)  - 1024^8

 ' store bytes
 Temp2# = Temp#
 ' 1.#INF and -1.#INF ad -1.#IND and 1#QNAN
 IF INSTR(STR$(Temp2#), "#") THEN
    ParseSize$ = LTRIM$(STR$(Temp2#))
    EXIT FUNCTION
 END IF
 ' zero byte
 IF Temp2# = 0# THEN
    ParseSize$ = "0 B."
    EXIT FUNCTION
 END IF
 ' store sign
 IF Temp2# < 0# THEN
    Sign = -1
    Temp2# = ABS(Temp2#)
 END IF
 ' reset type counter
 TempA = False
 ' divide by 1024 until less
 DO
    IF Temp2# >= 1024 THEN
       Temp2# = Temp2# / 1024#
       ' count type
       TempA = TempA + 1
       IF TempA = 9 THEN
	  EXIT DO
       END IF
    ELSE
       EXIT DO
    END IF
 LOOP
 ' check overflow
 IF TempA > 8 THEN
    VarX$ = LTRIM$(STR$(Temp#))
    Var1$ = LEFT$(VarX$, INSTR(VarX$, "D") - 1)
    Var2$ = MID$(VarX$, INSTR(VarX$, "D"))
    IF INSTR(Var1$, ".") THEN Var1$ = LEFT$(Var1$, INSTR(Var1$, ".") + 2)
    ParseSize$ = Var1$ + Var2$' 1.23D100
    EXIT FUNCTION
 END IF
 ' store type
 Strng2$ = Nul
 IF TempA >= 1 AND TempA <= 8 THEN ' powers of 1024
    Strng2$ = MID$("KMGTPEZY", TempA, 1)
 END IF
 ' return formatted string
 VarX$ = FORMAT$(Temp2#, "#,##0.00") + " " + Strng2$ + "B."
 IF Sign THEN
    VarX$ = "-" + VarX$
 END IF
 ParseSize$ = VarX$
END FUNCTION

' return PI double
FUNCTION PI#
  PI# = ATN(1) * 4
END FUNCTION

Sub Close.File
 CALL Get.Token3
 CALL Get.Token
 CALL Parse1(Temp#)
 Temp4# = Temp#
 Print Temp4#
 Do
    IF Strng = "," THEN
       CALL Get.Token
       CALL Parse1(Temp#)
       Temp4# = Temp#
       Print Temp4#
    Else
       Exit Do
    END IF
 Loop
 CALL Get.Token4
 Last.Token = False
End Sub

SUB Poke.Value
 CALL Get.Token3
 CALL Get.Token
 CALL Parse1(Temp#)
 Poke.Address = CINT(Temp#)
 CALL Get.Token
 CALL Parse1(Temp#)
 Poke.Byte = CINT(Temp#)
 POKE Poke.Address, Poke.Byte
 CALL Get.Token4
 Last.Token = True
END SUB

SUB RD.Directory
 CALL Get.Token3
 CALL Get.Token
 CALL Parse1(Temp#)
 CALL Get.Token4
 Hidden = True
 RMDIR Out3
 Last.Token = False
END SUB

' read lowercase command line from PSP
SUB Read.Command (Command.Line$)
Command.Line$ = Nul
InregsX.AX = &H6200
CALL InterruptX(&H21, InregsX, OutregsX)
PSPsegment = OutregsX.BX
PSPoffset = 128
DEF SEG = PSPsegment
FOR Count = 1 TO 127
   Command.Char = PEEK(PSPoffset + Count)
   SELECT CASE Command.Char
   CASE 0, 10, 13
      EXIT FOR
   CASE ELSE
      Command.Line$ = Command.Line$ + CHR$(Command.Char)
   END SELECT
NEXT
DEF SEG
END SUB

SUB Read.Rows (Var)
 ON LOCAL ERROR GOTO Error.Trap62x
 Temp1 = 50
 LOCATE 50, 1, 0
 Var = 50
 EXIT SUB
Next.Trap1:
 Temp1 = 43
 LOCATE 43, 1, 0
 Var = 43
 EXIT SUB
Next.Trap2:
 Temp1 = 25
 LOCATE 25, 1, 0
 Var = 25
 EXIT SUB
Next.Trap3:
 Var = 12
Error.Resume62x:
 EXIT SUB
Error.Trap62x:
 IF Temp1 = 50 THEN
    RESUME Next.Trap1
 END IF
 IF Temp1 = 43 THEN
    RESUME Next.Trap2
 END IF
 IF Temp1 = 25 THEN
    RESUME Next.Trap3
 END IF
 RESUME Error.Resume62x
END SUB

SUB Read.Stdin
 ' check stdin length
 InregsX.AX = &H4202 ' eof
 InregsX.BX = 0 ' stdin
 InregsX.CX = 0
 InregsX.DX = 0
 CALL InterruptX(&H21, InregsX, OutregsX)
 IF OutregsX.AX = 0 THEN ' no input found
    Var$ = INKEY$
    EXIT SUB
 END IF

 ' reset stdin start
 InregsX.AX = &H4200
 InregsX.BX = 0 ' stdin
 InregsX.CX = 0
 InregsX.DX = 0
 CALL InterruptX(&H21, InregsX, OutregsX)
 StdinEnabled = -1

 ' loop while reading input
 COLOR White, Black
 Out2 = Nul
 DO

    ' read character from standard input
    InregsX.AX = &H3F00
    InregsX.BX = 0 ' stdin
    InregsX.CX = 1 ' char
    InregsX.DS = VARSEG(Pipe.Buffer)
    InregsX.DX = VARPTR(Pipe.Buffer)
    CALL InterruptX(&H21, InregsX, OutregsX)
    IF (OutregsX.Flags AND &H1) = &H1 THEN
       EXIT DO
    END IF
    IF (OutregsX.Flags AND &H1) = &H0 THEN
       IF OutregsX.AX = 0 THEN
	  EXIT DO
       END IF
       Char$ = Pipe.Buffer

       ' determine character type
       IF (ASC(Char$) = 10 AND LINUX) OR (ASC(Char$) = 13 AND LINUX = 0) THEN
	  ' calculate line
	  Out2 = STRIM$(Out2)
	  IF LEN(Out2) THEN
	     CALL Strip.Semicolon(Out2)
	     Var$ = Out2
	     DO
		IF INSTR(Var$, ";") THEN
		   Out2 = LEFT$(Var$, INSTR(Var$, ";") - 1)
		   Var$ = MID$(Var$, INSTR(Var$, ";") + 1)
		   Out2 = LTRIM$(RTRIM$(Out2))
		   IF LEN(Out2) THEN
		      PRINT Out2; " ";
		      CALL Remove.Spaces
		      Recurse = 0
		      Hidden = False
		      CALL Enter.Equate
		   END IF
		   Out2 = Nul
		ELSE
		   Out2 = Var$
		   Out2 = LTRIM$(RTRIM$(Out2))
		   IF LEN(Out2) THEN
		      PRINT Out2; " ";
		      CALL Remove.Spaces
		      Recurse = 0
		      Hidden = False
		      CALL Enter.Equate
		      Out2 = Nul
		   END IF
		   EXIT DO
		END IF
	     LOOP
	  END IF
      ELSE ' store char
	 IF (ASC(Char$) = 10 AND LINUX = 0) THEN
	    Eat$ = ""
	 ELSE
	    Out2 = Out2 + Char$
	 END IF
      END IF
    END IF
 LOOP
 ' stop stdin
 COLOR 7, 0
 SYSTEM
END SUB

' remove all white spaces, skip white spaces in quotes.
SUB Remove.Spaces
Var$ = Out2
Temp = False
DO
    Temp = Temp + 1
    IF Temp > LEN(Var$) THEN
	EXIT DO
    END IF
    IF MID$(Var$, Temp, 1) = Quote$ THEN
	DO
	    Temp = Temp + 1
	    IF Temp > LEN(Var$) THEN
		EXIT DO
	    END IF
	    IF MID$(Var$, Temp, 1) = Quote$ THEN
		EXIT DO
	    END IF
	LOOP
    END IF
    FOR Blanks = 1 TO LEN(White.Space)
	IF MID$(Var$, Temp, 1) = MID$(White.Space, Blanks, 1) THEN
	    Var$ = LEFT$(Var$, Temp - 1) + MID$(Var$, Temp + 1)
	    Temp = Temp - 1
	    EXIT FOR
	END IF
    NEXT
LOOP
Out2 = Var$
END SUB

SUB Save.Screen2 (Var)
 ON LOCAL ERROR GOTO Error.Trap62
 IF Var = False THEN
    VarX3 = (Max.Row - 1) * 80 + 80
    REDIM Temp.ArrayY(VarX3) AS INTEGER
    REDIM Temp.ArrayZ(VarX3) AS INTEGER
    FOR Var1 = 1 TO Max.Row
       FOR Var2 = 1 TO 80
	  TempZ1 = SCREEN(Var1, Var2)
	  TempZ2 = SCREEN(Var1, Var2, 1)
	  VarX2 = (Var1 - 1) * 80 + Var2
	  Temp.ArrayY(VarX2) = TempZ1' char
	  Temp.ArrayZ(VarX2) = TempZ2' color
       NEXT
    NEXT
 ELSE
    CLS
    FOR VarX = Max.Row TO 1 STEP -1
       FOR VarY = 1 TO 80
	  VarX2 = (VarX - 1) * 80 + VarY
	  IF Temp.ArrayY(VarX2) <> 32 THEN
	     GOTO Next.Line
	  END IF
       NEXT
    NEXT
Next.Line:
    TempX = VarX
    FOR Var1 = 1 TO VarX' Max.Row
       FOR Var2 = 1 TO 80
	  VarX2 = (Var1 - 1) * 80 + Var2
	  VarB = INT(Temp.ArrayZ(VarX2) / 16)
	  VarF = Temp.ArrayZ(VarX2) MOD 16
	  TempZ1 = Temp.ArrayY(VarX2)
	  LOCATE Var1, Var2, 0
	  COLOR VarF, VarB
	  PRINT CHR$(TempZ1);
       NEXT
    NEXT
 END IF
Error.Resume62:
 EXIT SUB
Error.Trap62:
 RESUME Error.Resume62
END SUB

FUNCTION SearchLine$ (VarZ%)
 Var1 = 0
 FOR Var2 = Max.History TO 1 STEP -1
     IF RTRIM$(History2(Temp0, Var2)) <> Nul THEN
	 Var1 = Var2
	 EXIT FOR
     END IF
 NEXT
 IF Var1 = 0 THEN
    PRINT "Text not found."
    EXIT FUNCTION
 END IF
 IF Var1 > 0 THEN
    COLOR White
    Var2$ = "Enter search text(? and * allowed):"
    PRINT Var2$;
    LINE INPUT Ascii$
    COLOR Yellow
    IF LEN(Ascii$) THEN
       FOR V = 1 TO Var1
	  V1$ = LCASE$(History2(Temp0, V))
	  V2$ = LCASE$(Ascii$)
	  CALL InstrSUB(Var, V1$, V2$)
	  IF Var THEN
	     Var$ = RTRIM$(History2(Temp0, V))
	     SearchLine$ = Var$
	     PRINT ">";
	     FOR VarX = 1 TO LEN(Var$)
		PRINT MID$(Var$, VarX, 1);
	     NEXT
	     EXIT FUNCTION
	  END IF
       NEXT
    END IF
    PRINT "Text not found.";
 END IF
END FUNCTION

' display statusline
'   init max.row
SUB StatusLine
IF StdinEnabled THEN EXIT SUB
LOCATE Max.Row, 1, 1
COLOR White, Blue
S$ = "Whatis utility " + Version$ + " " + Release$ + " " + FORMAT$(Now, "mm-dd-yyyy") + " " + FORMAT$(Now, "hh:mm:ss")
S$ = S$ + SPACE$(80 - LEN(S$))
PRINT S$;
LOCATE Max.Row - 1, 1, 1
COLOR White, Black
END SUB

SUB Swap.Data
 Temp1$ = MID$(Out2, 5)
 Temp2$ = UCASE$(LEFT$(Temp1$, 1))
 IF Temp2$ >= "A" AND Temp2$ <= "Z" THEN
    Variable1 = ASC(Temp2$) - 64
    Temp1$ = MID$(Temp1$, 2)
    IF LEFT$(Temp1$, 1) = "," THEN
       Temp1$ = MID$(Temp1$, 2)
       Temp2$ = UCASE$(LEFT$(Temp1$, 1))
       IF Temp2$ >= "A" AND Temp2$ <= "Z" THEN
	  Variable2 = ASC(Temp2$) - 64
	  Temp1$ = MID$(Temp1$, 2)
	  IF Temp1$ = Nul THEN
	     SWAP Variables(Variable1), Variables(Variable2)
	     EXIT SUB
	  ELSE
	     Var1$ = LEFT$(Temp1$, 1)
	     IF Var1$ = "(" OR Var1$ = "[" OR Var1$ = "{" THEN
		Out2 = MID$(Temp1$, 2)
		Last.Token = False
		Token.Index = 1
		CALL Get.Token
		CALL Parse1(Temp3#)
		Element = CINT(Temp3#)
		SELECT CASE Var1$
		CASE "("
		   IF Strng <> ")" THEN
		      Strng = "<bad closing token>"
		      ERROR 92
		   END IF
		CASE "["
		   IF Strng <> "]" THEN
		      Strng = "<bad closing token>"
		      ERROR 92
		   END IF
		CASE "{"
		   IF Strng <> "}" THEN
		      Strng = "<bad closing token>"
		      ERROR 92
		   END IF
		END SELECT
		IF Element > False AND Element <= Max.Arrays THEN
		   SWAP Variables(Variable1), Arrays(Variable2, Element)
		   EXIT SUB
		END IF
	     END IF
	  END IF
       END IF
    ELSE
       Var1$ = LEFT$(Temp1$, 1)
       IF Var1$ = "(" OR Var1$ = "[" OR Var1$ = "{" THEN
	  Out2 = MID$(Temp1$, 2)
	  Last.Token = False
	  Token.Index = 1
	  CALL Get.Token
	  CALL Parse1(Temp3#)
	  SELECT CASE Var1$
	  CASE "("
	     IF Strng <> ")" THEN
		Strng = "<bad closing token>"
		ERROR 92
	     END IF
	  CASE "["
	     IF Strng <> "]" THEN
		Strng = "<bad closing token>"
		ERROR 92
	     END IF
	  CASE "{"
	     IF Strng <> "}" THEN
		Strng = "<bad closing token>"
		ERROR 92
	     END IF
	  END SELECT
	  Element1 = CINT(Temp3#)
	  Temp1$ = MID$(Out2, Token.Index)
	  IF LEFT$(Temp1$, 1) = "," THEN
	     Out2 = MID$(Temp1$, 2)
	     Out2 = UCASE$(Out2)
	     Temp1$ = MID$(Temp1$, 2)
	     Temp2$ = UCASE$(LEFT$(Temp1$, 1))
	     IF Temp2$ >= "A" AND Temp2$ <= "Z" THEN
		Variable2 = ASC(Temp2$) - 64
		Temp1$ = MID$(Temp1$, 2)
		IF Temp1$ = Nul THEN
		   IF Element1 > False AND Element1 <= Max.Arrays THEN
		      SWAP Arrays(Variable1, Element1), Variables(Variable2)
		      EXIT SUB
		   END IF
		ELSE
		   Var1$ = LEFT$(Temp1$, 1)
		   IF Var1$ = "(" OR Var1$ = "[" OR Var1$ = "{" THEN
		      Out2 = MID$(Temp1$, 2)
		      Last.Token = False
		      Token.Index = 1
		      CALL Get.Token
		      CALL Parse1(Temp3#)
		      SELECT CASE Var1$
		      CASE "("
			 IF Strng <> ")" THEN
			    Strng = "<bad closing token>"
			    ERROR 92
			 END IF
		      CASE "["
			IF Strng <> "]" THEN
			    Strng = "<bad closing token>"
			    ERROR 92
			 END IF
		      CASE "{"
			 IF Strng <> "}" THEN
			    Strng = "<bad closing token>"
			    ERROR 92
			 END IF
		      END SELECT
		      Element2 = CINT(Temp3#)
		      IF Element1 > False AND Element1 <= Max.Arrays THEN
			 IF Element2 > False AND Element2 <= Max.Arrays THEN
			    SWAP Arrays(Variable1, Element1), Arrays(Variable2, Element2)
			    EXIT SUB
			 END IF
		      END IF
		   END IF
		END IF
	     END IF
	  END IF
       ELSE
	  IF LEFT$(Temp1$, 1) = "$" THEN
	     Temp1$ = MID$(Temp1$, 2)
	     IF LEFT$(Temp1$, 1) = "," THEN
		Temp1$ = MID$(Temp1$, 2)
		Temp2$ = UCASE$(LEFT$(Temp1$, 1))
		IF Temp2$ >= "A" AND Temp2$ <= "Z" THEN
		   Variable2 = ASC(Temp2$) - 64
		   Temp1$ = MID$(Temp1$, 2)
		   IF Temp1$ = "$" THEN
		      SWAP Strngs(Variable1), Strngs(Variable2)
		      EXIT SUB
		   END IF
		END IF
	     END IF
	  END IF
       END IF
    END IF
 END IF
 ERROR 150
END SUB

