REM Freespace for Whatis.

' start Qbx /l Qbx.Qlb
REM $DYNAMIC
DEFINT A-Z

REM file: FAT32.BI - FAT32 structure

TYPE FAT32Type
 Size1 AS STRING * 2
 Version AS STRING * 2
 SectorsPerCluster AS STRING * 4
 BytesPerSector AS STRING * 4
 FreeClusters AS STRING * 4
 TotalClusters AS STRING * 4
 FreeSectors AS STRING * 4
 TotalSectors AS STRING * 4
 FreeUnits AS STRING * 4
 TotalUnits AS STRING * 4
 Reserved AS STRING * 8
END TYPE

DECLARE SUB FreeSpace (T%, T$)

TYPE RegTypeX
 AX AS INTEGER
 BX AS INTEGER
 CX AS INTEGER
 DX AS INTEGER
 BP AS INTEGER
 SI AS INTEGER
 DI AS INTEGER
 Flags AS INTEGER
 DS AS INTEGER
 ES AS INTEGER
END TYPE

DECLARE SUB InterruptX (N AS INTEGER, I AS RegTypeX, O AS RegTypeX)

DIM SHARED InregsX AS RegTypeX, OutregsX AS RegTypeX

' get drive freespace (modified 12/09/2022)
' input: T% = 1  freespace
'        T% = 2  totalspace
'        T% = 3  usedspace
' output: T$ = diskspace
SUB FreeSpace (T%, T$)
 ON LOCAL ERROR GOTO DriveError
 DIM ASCIIZ AS STRING * 260

 ' declare structures
 DIM FAT32Struc AS FAT32Type

 ' get drive letter
 T# = 0#
 D$ = LEFT$(T$, 1)
 D$ = UCASE$(D$)
 IF D$ >= "A" AND D$ <= "Z" THEN
    V = ASC(D$) - 64
 ELSE
    V = 3
 END IF

 ' get FAT32 drive space
 ASCIIZ = CHR$(V + 64) + ":\" + CHR$(0)
 InregsX.AX = &H7303
 InregsX.DS = VARSEG(ASCIIZ)
 InregsX.DX = VARPTR(ASCIIZ)
 InregsX.ES = VARSEG(FAT32Struc)
 InregsX.DI = VARPTR(FAT32Struc)
 InregsX.CX = LEN(FAT32Struc)
 CALL InterruptX(&H21, InregsX, OutregsX)
 IF (OutregsX.Flags AND &H1) = &H0 THEN
    IF T% = 1 THEN ' free
       GOSUB Get.Free
    END IF
    IF T% = 2 THEN ' total
       GOSUB Get.Total
    END IF
    IF T% = 3 THEN ' used
       GOSUB Get.Total
       Total# = T#
       GOSUB Get.Free
       Free# = T#
       T# = Total# - Free#
    END IF
    IF T# THEN
       T$ = LTRIM$(STR$(T#))
       EXIT SUB
    END IF
 END IF

 ' check disk space.
 InregsX.AX = &H3600
 InregsX.DX = V
 CALL InterruptX(&H21, InregsX, OutregsX)
 ' check drive exists.
 IF OutregsX.AX = &HFFFF THEN
    T# = 0#
 ELSE
    ' Note: FAT16 only reports up to 2 GB.
    IF T% = 1 THEN ' free
       GOSUB Get.Free16
    END IF
    IF T% = 2 THEN ' total
       GOSUB Get.Total16
    END IF
    IF T% = 3 THEN ' used
       GOSUB Get.Total16
       Total# = T#
       GOSUB Get.Free16
       Free# = T#
       T# = Total# - Free#
    END IF
 END IF
DriveResume:
 IF T# = 0# THEN
    T# = 256 * 1024 ^ 2
 END IF
 T$ = LTRIM$(STR$(T#))
 EXIT SUB
Get.Free:
 Bytes# = CDBL(ASC(MID$(FAT32Struc.BytesPerSector, 1, 1)))
 Bytes# = Bytes# + CDBL(ASC(MID$(FAT32Struc.BytesPerSector, 2, 1))) * 256#
 Bytes# = Bytes# + CDBL(ASC(MID$(FAT32Struc.BytesPerSector, 3, 1))) * 65536#
 Bytes# = Bytes# + CDBL(ASC(MID$(FAT32Struc.BytesPerSector, 4, 1))) * 16777216#
 Sectors# = CDBL(ASC(MID$(FAT32Struc.FreeSectors, 1, 1)))
 Sectors# = Sectors# + CDBL(ASC(MID$(FAT32Struc.FreeSectors, 2, 1))) * 256#
 Sectors# = Sectors# + CDBL(ASC(MID$(FAT32Struc.FreeSectors, 3, 1))) * 65536#
 Sectors# = Sectors# + CDBL(ASC(MID$(FAT32Struc.FreeSectors, 4, 1))) * 16777216#
 T# = Bytes# * Sectors#
 RETURN
Get.Total:
 Bytes# = CDBL(ASC(MID$(FAT32Struc.BytesPerSector, 1, 1)))
 Bytes# = Bytes# + CDBL(ASC(MID$(FAT32Struc.BytesPerSector, 2, 1))) * 256#
 Bytes# = Bytes# + CDBL(ASC(MID$(FAT32Struc.BytesPerSector, 3, 1))) * 65536#
 Bytes# = Bytes# + CDBL(ASC(MID$(FAT32Struc.BytesPerSector, 4, 1))) * 16777216#
 Sectors# = CDBL(ASC(MID$(FAT32Struc.TotalSectors, 1, 1)))
 Sectors# = Sectors# + CDBL(ASC(MID$(FAT32Struc.TotalSectors, 2, 1))) * 256#
 Sectors# = Sectors# + CDBL(ASC(MID$(FAT32Struc.TotalSectors, 3, 1))) * 65536#
 Sectors# = Sectors# + CDBL(ASC(MID$(FAT32Struc.TotalSectors, 4, 1))) * 16777216#
 T# = Bytes# * Sectors#
 RETURN
Get.Free16:
 ' free space on drive in bytes is AX * BX * CX adjusted signed-int
 T1# = OutregsX.AX: IF T1# < 0# THEN T1# = T1# + 65536#
 T2# = OutregsX.BX: IF T2# < 0# THEN T2# = T2# + 65536#
 T3# = OutregsX.CX: IF T3# < 0# THEN T3# = T3# + 65536#
 T# = T1# * T2# * T3#
 RETURN
Get.Total16:
 ' total space on drive in bytes is AX * CX * DX adjusted signed-int
 T1# = OutregsX.AX: IF T1# < 0# THEN T1# = T1# + 65536#
 T2# = OutregsX.CX: IF T2# < 0# THEN T2# = T2# + 65536#
 T3# = OutregsX.DX: IF T3# < 0# THEN T3# = T3# + 65536#
 T# = T1# * T2# * T3#
 RETURN
DriveError:
 RESUME DriveResume
END SUB

